#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

import os
cfg = {}  # Will be updated by CK (meta description of this module)
work = {}  # Will be updated by CK (temporal data)
ck = None  # Will be updated by CK (initialized CK kernel)

# Local settings


##############################################################################
# Initialize module


def init(i):
    """
    Input:  {}

    Output: {
              return   - return code =  0, if successful
                                     >  0, if error
              (error)  - error text if return > 0
            }

    """
    return {'return': 0}

##############################################################################
# Create repository in a given directory and record info in CK


def add(i):
    """
    Note, that we can't create repos in parallel (recording to repo cache may fail).
    However, for now, we do not expect such cases (i.e. repos are created rarely)

    Input:  {
              (repo_uoa)                 - repo UOA (where to create entry)
              data_uoa                   - data UOA
              (data_uid)                 - data UID (if uoa is an alias)
              (data_name)                - user friendly data name

              (cids[0])                  - as uoa or full CID

              (path)                     - if !='' - create in this path or import from this path
              (here)                     - if =='yes', create in current path
              (use_default_path)         - if 'yes' create repository in the default path (CK_REPOS)
                                           instead of the current path (default is 'yes')

              (use_current_path)         - if 'yes' create repository in the current path
                                           (default is 'no')

              (default)                  - if 'yes', no path is used, 
                                           but the repository is taken either 
                                           from the CK directory or from CK_LOCAL_REPO

              (import)                   - if 'yes', register repo in the current directory in CK
                                           (when received from someone else)

              (remote)                   - if 'yes', remote repository
              (remote_repo_uoa)          - if !='' and type=='remote' repository UOA on the remote CK server

              (shared)                   - if not remote and =='git', repo is shared/synced through GIT
              (share)                    - (for user-friendly CMD) if 'yes', set shared=git

              (allow_writing)            - if 'yes', allow writing 
                                           (useful when kernel is set to allow writing only to such repositories)

              (url)                      - if type=='remote' or 'git', URL of remote repository or git repository
              (hostname)                 - if !='', automatically form url above (add http:// + /ck?)
              (port)                     - if !='', automatically add to url above
              (hostext)                  - if !='', add to the end of above URL instead of '/ck?' -
                                           useful when CK server is accessed via Apache2, IIS, Nginx or other web servers

              (githubuser)               - if shared repo, use this GitHub user space instead of default "ctuning"
              (sync)                     - if 'yes' and type=='git', sync repo after each write operation

              (gitzip)                   - if 'yes', download as zip from GitHub
              (zip)                      - path to zipfile (local or remote http/ftp)
              (overwrite)                - if 'yes', overwrite files when unarchiving

              (repo_deps)                - dict with dependencies on other shared repositories with following keys:
                                             "repo_uoa"
                                             ("repo_uid") - specific UID (version) of a repo
                                             ("repo_url") - URL of the shared repository (if not from github.com/ctuning)

              (quiet)                    - if 'yes', do not ask questions unless really needed

              (skip_reusing_remote_info) - if 'yes', do not reuse remote .cmr.json description of a repository

              (current_repos)            - if resolving dependencies on other repos, list of repos being updated (to avoid infinite recursion)

              (describe)                 - describe repository for Artifact Evaluation (see http://cTuning.org/ae)

              (stable)        - take stable version (highly experimental)
              (version)       - checkout version (default - stable)
              (branch)        - git branch
              (checkout)      - git checkout

              (private)                  - if 'yes', mark as private (do not automatically list entries, etc)

              (split_all_dirs)           - if !='0' force split of all dirs in this repo (be careful)
                                           must be empty before doing this!

              (recache)                  - if 'yes' force recache
            }

    Output: {
              return       - return code =  0, if successful
                                           16, repository with a given path is already registered in CK
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    # Check if global writing is allowed
    r = ck.check_writing({})
    if r['return'] > 0:
        return r

    o = i.get('out', '')

    a = i.get('repo_uoa', '')
    d = i.get('data_uoa', '')
    di = i.get('data_uid', '')
    dn = i.get('data_name', '')

    if ck.cfg.get('allowed_entry_names', '') != '':
        import re

        anames = ck.cfg.get('allowed_entry_names', '')

        if not re.match(anames, a) or \
           not re.match(anames, d) or \
           not re.match(anames, di):
            return {'return': 1, 'error': 'found disallowed characters in names (allowed: "'+anames+'")'}

    if ck.cfg.get('force_lower', '') == 'yes':
        a = a.lower()
        d = d.lower()
        di = di.lower()

    stable = i.get('stable', '')
    version = i.get('version', '')
    if stable == 'yes':
        version = 'stable'
    branch = i.get('branch', '')
    checkout = i.get('checkout', '')

    cr = i.get('current_repos', [])

    rdeps = i.get('repo_deps', [])

    quiet = i.get('quiet', '')

    overwrite = i.get('overwrite', '')

    remote = i.get('remote', '')
    rruoa = i.get('remote_repo_uoa', '')
    shared = i.get('shared', '')
    if shared == 'yes':
        shared = 'git'

    share = i.get('share', '')
    if share == 'yes' and shared == '':
        shared = 'git'

    private = i.get('private', '')
    split_all_dirs = i.get('split_all_dirs', '')
    xrecache = i.get('recache', '')

    ptr = ''
    zp = i.get('zip', '')
    gz = i.get('gitzip', '')

    if zp != '':
        if zp.startswith('~'):
            from os.path import expanduser
            home = expanduser("~")

            zp = os.path.abspath(home+os.sep+zp[1:])

        quiet = 'yes'

        if d == '':
            # Try to get data UOA
            if not os.path.isfile(zp):
                return {'return': 1, 'error': 'zip file not found'}

            # Try to get .ckr.json
            import zipfile

            try:
                with zipfile.ZipFile(zp) as z:
                    zip_has_git = False
                    if '.git/HEAD' in z.namelist():
                        zip_has_git = True

                    if ck.cfg['repo_file'] in z.namelist():
                        x = z.open(ck.cfg['repo_file'])
                        y = x.read()

                        r = ck.convert_json_str_to_dict(
                            {'str': y, 'skip_quote_replacement': 'yes'})
                        if r['return'] > 0:
                            return r
                        yd = r['dict']

                        d = yd.get('data_uoa', '')
                        di = yd.get('data_uid', '')
                        dn = yd.get('data_name', '')

                        zip_shared = yd.get('dict', {}).get('shared', '')
                        if zip_shared != '':
                            if zip_shared != 'git' or zip_has_git:
                                shared = yd['dict']['shared']
                                share = 'yes'

                        x.close()
                z.close()
            except Exception as e:
                return {'return': 1, 'error': 'problem reading zip file ('+format(e)+')'}

            if d == '':
                x1 = os.path.basename(zp)
                d = os.path.splitext(x1)[0]
                if d.startswith('ckr-'):
                    d = d[4:]

                if o == 'con':
                    ck.out('Auto-detected repo name from zip filename: '+d)

    if gz == 'yes':
        zp = ck.cfg['default_shared_repo_url']+'/'+d+'/archive/master.zip'
        ptr = d+'-master/'
        quiet = 'yes'

    rx = form_url(i)
    if rx['return'] > 0:
        return rx
    url = rx['url']

    sync = i.get('sync', '')
    df = i.get('default', '')

    eaw = i.get('allow_writing', '')

    udp = i.get('use_default_path', 'yes')
    ucp = i.get('use_current_path', '')
    if ucp == 'yes':
        udp = ''

    # Get repo path (unless 'here' later)
    px = i.get('path', '')

    # Check if import
    imp = i.get('import', '')
    if imp == 'yes':
        if px == '':
            i['here'] = 'yes'

    # Get 'here' path
    if i.get('here', '') == 'yes':
        px = os.getcwd()
    p = px

    if imp == 'yes':
        py = os.path.join(p, ck.cfg['repo_file'])
        if os.path.isfile(py):
            r = ck.load_json_file({'json_file': py})
            if r['return'] > 0:
                return r
            dc = r['dict']

            d = dc.get('data_uoa', '')
            di = dc.get('data_uid', '')
            dn = dc.get('data_name', '')

    if p == '' and udp == 'yes':
        p = os.path.join(ck.work['dir_repos'], d)

    # Normalize path
    p = os.path.normpath(p)

    # If console mode, first, check if shared (GIT, etc)
    if o == 'con':
        # Asking for alias
        if df != 'yes' and (d == '' or ck.is_uid(d)):
            r = ck.inp(
                {'text': 'Enter an alias for this repository (or Enter to skip it): '})
            d = r['string']
            if d == '':
                d = di
            if d == '':
                r = ck.gen_uid({})
                if r['return'] > 0:
                    return r
                di = r['data_uid']
                d = di

        # Asking for a user-friendly name
        if df != 'yes' and dn == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Enter a user-friendly name of this repository (or Enter to reuse alias): '})
                dn = r['string']
            if dn == '':
                dn = d

        # Asking if remote
        if df != 'yes' and remote == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Is this repository a remote CK web service (y/N)? '})
                remote = r['string'].lower()
            if remote == 'yes' or remote == 'y':
                remote = 'yes'
            else:
                remote = ''

        # Asking for a user-friendly name
        if px == '' and df != 'yes' and remote != 'yes' and udp == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Would you like to create repo in the directory from CK_REPOS variable (Y/n): '})
                x = r['string'].lower()
                if x == '' or x == 'yes' or x == 'y':
                    p = os.path.join(ck.work['dir_repos'], d)

        # Asking for remote url
        if df != 'yes' and remote == 'yes' and url == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Enter URL of remote CK repo (example: http://localhost:3344/ck?): '})
                url = r['string'].lower()
            if url == '':
                return {'return': 1, 'error': 'URL is empty'}

        # Asking for remote repo UOA
        if df != 'yes' and remote == 'yes' and rruoa == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Enter remote repo UOA or Enter to skip: '})
                rruoa = r['string'].lower()

        # Asking for shared
        if remote == '' and shared == '' and share == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Is this repository shared via GIT (y/N)? '})
                x = r['string'].lower()
                if x == 'yes' or x == 'y':
                    share = 'yes'

        if share == 'yes' and shared == '':
            shared = 'git'

        # Check additional parameters if git
        ghu = i.get('githubuser', '')
        if shared == 'git' and url == '':

            if ghu != '':
                durl = ck.cfg.get('github_repo_url', '')
                if not durl.endswith('/'):
                    durl += '/'
                durl += ghu
            else:
                durl = ck.cfg.get('default_shared_repo_url', '')

            if not durl.endswith('/'):
                durl += '/'
            durl += d
#          if durl.startswith('http://') or durl.startswith('https://'):
#             durl+='.git'

            if quiet != 'yes':
                s = 'Enter URL of GIT repo '
                if d == '':
                    s += '(for example, https://github.com/ctuning/ck-analytics.git)'
                else:
                    s += '(or Enter for '+durl+')'
                r = ck.inp({'text': s+': '})
                url = r['string'].lower()
            if url == '':
                url = durl

        # Check additional parameters if git
        if shared == 'git' and sync == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Would you like to sync repo each time after writing to it (y/N)?: '})
                x = r['string'].lower()
                if x == 'yes' or x == 'y':
                    sync = 'yes'

        # Asking for a user-friendly name
        if df != 'yes' and remote != 'yes' and eaw == '':
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Would you like to explicitly allow writing to this repository in case kernel disables all writing (y/N): '})
                x = r['string'].lower()
                if x == 'yes' or x == 'y':
                    eaw = 'yes'

        # Check if add more deps
        if quiet != 'yes':
            r = add_more_deps({})
            if r['return'] > 0:
                return r

            rdeps1 = r['repo_deps']
            for q in rdeps1:
                rdeps.append(q)

    # Check if already registered (if not remote)
    if remote != 'yes':
        r = ck.find_repo_by_path({'path': p})
        if r['return'] > 0 and r['return'] != 16:
            return r

    # Check if repository is already registered with this path
    r = ck.find_repo_by_path({'path': p})
    if r['return'] == 0:
        return {'return': 16, 'error': 'repository with a given path is already registered in CK'}
    elif r['return'] != 16:
        return r

    # Prepare local description file
    py = os.path.join(p, ck.cfg['repo_file'])

    # Create dummy if doesn't exist
    if remote != 'yes' and not os.path.isdir(p):
        os.makedirs(p)

    # If zip, get (download) and unzip file ...
    if zp != '':
        rz = get_and_unzip_archive(
            {'zip': zp, 'path': p, 'path_to_remove': ptr, 'overwrite': overwrite, 'out': o})
        if rz['return'] > 0:
            return rz

        py = os.path.join(p, ck.cfg['repo_file'])
        if os.path.isfile(py):
            r = ck.load_json_file({'json_file': py})
            if r['return'] > 0:
                return r
            dc = r['dict']

            d = dc.get('data_uoa', '')
            di = dc.get('data_uid', '')
            dn = dc.get('data_name', '')

    # If git, clone repo
    repo_had_local = True
    dd = {}
    if remote != 'yes' and shared == 'git' and zp == '':
        r = pull({'path': p, 'type': shared, 'url': url,
                  'clone': 'yes', 'git': i.get('git', ''), 'out': o})
        if r['return'] > 0:
            return r

        # Check if there is a local repo description
        if os.path.isfile(py):
            r = ck.load_json_file({'json_file': py})
            if r['return'] > 0:
                return r
            dc = r['dict']
            ddc = dc.get('dict', {})
            dd.update(ddc)

            xd = dc.get('data_uoa', '')
            xdi = dc.get('data_uid', '')
            xdn = dc.get('data_name', '')

            if o == 'con':
                ck.out('Cloned repository has the following info:')
                ck.out(' UID                = '+xdi)
                ck.out(' UOA                = '+xd)
                ck.out(' User friendly name = '+xdn)
                ck.out('')
            if i.get('skip_reusing_remote_info', '') != 'yes':
                d = xd
                di = xdi
                dn = xdn
        else:
            repo_had_local = False

    # Prepare meta description
    if df == 'yes':
        dd['default'] = 'yes'
    if remote == 'yes':
        dd['remote'] = 'yes'
        if rruoa != '':
            dd['remote_repo_uoa'] = rruoa
    if shared != '':
        dd['shared'] = shared
        if sync != '':
            dd['sync'] = sync
    if url != '':
        dd['url'] = url
    if remote != 'yes':
        dd['path'] = p
    if eaw == 'yes':
        dd['allow_writing'] = 'yes'
    if len(rdeps) > 0:
        dd['repo_deps'] = rdeps
    if private != '':
        dd['private'] = private
    if split_all_dirs != '':
        dd['split_all_dirs'] = split_all_dirs
    if xrecache != '':
        dd['recache'] = xrecache

    # Check if need to describe for Artifact Evaluation
    if i.get('describe', '') == 'yes':
        r = describe({'dict': dd})
        if r['return'] > 0:
            return r

    # If not default, go to common core function to create entry
    if df != 'yes':
        ii = {'module_uoa': work['self_module_uoa'],
              'action': 'add',
              'data_uoa': d,
              'dict': dd,
              'common_func': 'yes'}
        if a != '':
            ii['repo_uoa'] = a
        if di != '':
            ii['data_uid'] = di
        if dn != '':
            ii['data_name'] = dn
        rx = ck.access(ii)
        if rx['return'] > 0:
            return rx
    else:
        # Load default repo and prepare return
        ii = {'module_uoa': work['self_module_uoa'],
              'action': 'load',
              'data_uoa': 'default',
              'common_func': 'yes'}
        rx = ck.access(ii)
    px = rx['path']
    dx = rx['data_uid']
    alias = rx['data_alias']

    # Update repo cache if not default local
    dz = {'data_uoa': d, 'data_uid': dx, 'data_alias': alias,
          'path_to_repo_desc': px, 'data_name': dn, 'dict': dd}

    if df != 'yes':
        r = ck.reload_repo_cache({})  # Ignore errors
        ck.cache_repo_uoa[d] = dx
        ck.cache_repo_info[dx] = dz
        r = ck.save_repo_cache({})
        if r['return'] > 0:
            return r

    # Record local info of the repo (just in case)
    if remote != 'yes':
        if 'path_to_repo_desc' in dz:
            # Avoid recording some local info
            del (dz['path_to_repo_desc'])
        if dz.get('dict', {}).get('path', '') != '':
            del (dz['dict']['path'])  # Avoid recording some local info

        if not os.path.isfile(py):
            ry = ck.save_json_to_file({'json_file': py, 'dict': dz})
            if ry['return'] > 0:
                return ry

        # If sync (or pulled repo did not have local description), add it ...
        if sync == 'yes' or (shared == 'git' and not repo_had_local):
            ppp = os.getcwd()

            os.chdir(p)
            ss = ck.cfg['repo_types'][shared]['add'].replace(
                '$#path#$', px).replace('$#files#$', ck.cfg['repo_file'])
            os.system(ss)

            os.chdir(ppp)

    # If console mode, print various info
    if o == 'con':
        ck.out('')
        ck.out('CK repository successfully registered!')
        ck.out('')

        if df != 'yes':
            ck.out('CK repo description path = '+px)
            ck.out('CK repo UID              = '+dx)

    # Recache repos otherwise may be problems with deps
    if o == 'con':
        ck.out('')
        ck.out('Recaching repos to speed up access ...')
        ck.out('')
    r = recache({'out': o})
    if r['return'] > 0:
        return r

    # Check deps
    if o == 'con':
        ck.out('  ========================================')
        ck.out('  Checking dependencies on other repos ...')
        ck.out('')

    how = 'pull'
    if gz == 'yes':
        how = 'add'
    r = deps({'path': p,
              'current_path': cr,
              'how': how,
              'version': version,
              'branch': branch,
              'checkout': checkout,
              'out': o})
    if r['return'] > 0:
        return r

    # Print if default
    if df == 'yes' and o == 'con':
        ck.out('')
        ck.out('Please, do not forget to add path to this repository to CK_LOCAL_REPO environment variable:')
        ck.out('')
        ck.out('  Linux: export CK_LOCAL_REPO='+p)
        ck.out('  Windows: set CK_LOCAL_REPO='+p)

    return rx

##############################################################################
# Update repository in a given directory and record info in CK


def update(i):
    """
    Update repository info

    Input:  {
              data_uoa                   - data UOA of the repo

              (shared)                   - if not remote and =='git', shared through GIT

              (url)                      - if type=='git', URL of remote repository or git repository
              (hostname)                 - if !='', automatically form url above (add http:// + /ck?)
              (port)                     - if !='', automatically add to url above
              (hostext)                  - if !='', add to the end of above URL instead of '/ck?' -
                                           useful when CK server is accessed via Apache2, IIS, Nginx or other web servers

              (sync)                     - if 'yes' and type=='git', sync repo after each write operation
              (allow_writing)            - if 'yes', allow writing 
                                           (useful when kernel is set to allow writing only to such repositories)

              (repo_deps)                - dict with dependencies on other shared repositories with following keys:
                                             "repo_uoa"
                                             ("repo_uid") - specific UID (version) of a repo
                                             ("repo_url") - URL of the shared repository (if not from github.com/ctuning)

              (update)                   - if 'yes', force updating

              (describe)                 - describe repository for Artifact Evaluation (see http://cTuning.org/ae)

              (private)                  - if 'yes', mark as private (do not automatically list entries, etc)

              (split_all_dirs)           - if !='0' force split of all dirs in this repo (be careful)
                                           must be empty before doing this!

              (recache)                  - if 'yes' force recache

              (quiet)                    - if 'yes', do not answer extra questions

              (dict)                     - update dict directly
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy

    # Check if global writing is allowed
    r = ck.check_writing({})
    if r['return'] > 0:
        return r

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    remote = i.get('remote', '')
    rruoa = i.get('remote_repo_uoa', '')
    shared = i.get('shared', '')
    sync = i.get('sync', '')

    rx = form_url(i)
    if rx['return'] > 0:
        return rx
    url = rx['url']

    rdeps = i.get('repo_deps', [])

    eaw = i.get('allow_writing', '')

    quiet = i.get('quiet', '')

    # Get configuration (not from Cache - can be outdated info!)
#    r=ck.load_repo_info_from_cache({'repo_uoa':duoa})
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uoa'],
                   'data_uoa': duoa})
    if r['return'] > 0:
        return r

    p = r.get('dict', {}).get('path', '')

    dn = r.get('data_name', '')
    d = r['dict']

    dd = i.get('dict', {})
    if len(dd) > 0:
        d.update(dd)

    remote = d.get('remote', '')

    changed = False

    # Check user-friendly name
    if dn != '':
        ck.out('Current user-friendly name of this repository: '+dn)
        ck.out('')

        if quiet != 'yes':
            r = ck.inp(
                {'text': 'Enter a user-friendly name of this repository (or Enter to keep old value): '})
            x = r['string']
            if x != '':
                dn = x
                changed = True

    # If remote or shared, update URL
    if shared == 'yes' or shared == 'git':
        d['shared'] = 'git'

    shared = d.get('shared', '')
    if remote == 'yes':
        url = d.get('url', '')
        ck.out('Repository is remote ...')
        ck.out('')
        ck.out('Current URL: '+url)
        ck.out('')
        if quiet != 'yes':
            rx = ck.inp(
                {'text': 'Enter new URL (or Enter to leave old one): '})
            x = rx['string']
            if x != '':
                d['url'] = x
                changed = True
    elif shared != '':
        if url != '':
            d['url'] = url

        url = d.get('url', '')
        ck.out('Repository is shared ...')
        ck.out('')
        ck.out('Current URL: '+url)

        if shared == 'git':
            sync = d.get('sync', '')
            ck.out('')
            if sync != '':
                ck.out('Current sync setting: '+sync)
            if quiet != 'yes':
                r = ck.inp(
                    {'text': 'Would you like to sync repo each time after writing to it (y/N)?: '})
                x = r['string'].lower()
                if x == 'yes' or x == 'y':
                    d['sync'] = 'yes'
                    changed = True

    # Asking about forbidding explicit writing to this repository
    if remote != 'yes' and eaw == '':
        if eaw == '':
            eaw = d.get('allow_writing', '')
        ck.out('')
        if eaw != '':
            ck.out('Current "allow writing" setting: '+eaw)

        if quiet != 'yes':
            r = ck.inp(
                {'text': 'Would you like to allow explicit writing to this repository when kernel disables all writing (y/N): '})
            x = r['string'].lower()
            if x == 'yes' or x == 'y':
                d['allow_writing'] = 'yes'
                changed = True

    # Check if explicit deps
    if len(rdeps) > 0:
        if 'repo_deps' not in d:
            d['repo_deps'] = rdeps
        else:
            for q in rdeps:
                d['repo_deps'].append(q)
        changed = True

    # Print deps
    rdeps = d.get('repo_deps', [])
    if len(rdeps) > 0:
        ck.out('')
        ck.out('Current dependencies on other repositories:')
        r = print_deps({'repo_deps': rdeps, 'out': o, 'out_prefix': '  '})
        if r['return'] > 0:
            return r
        ck.out('')

    # Check if add more deps
    if quiet != 'yes':
        r = add_more_deps({})
        if r['return'] > 0:
            return r

        rdeps1 = r['repo_deps']
        if len(rdeps1) > 0:
            if 'repo_deps' not in d:
                d['repo_deps'] = rdeps1
            else:
                for q in rdeps1:
                    d['repo_deps'].append(q)
            changed = True

    # Check if need to describe for Artifact Evaluation
    if i.get('describe', '') == 'yes':
        r = describe({'dict': d})
        if r['return'] > 0:
            return r
        changed = True

    # Check if private
    private = i.get('private', '')
    if private != '':
        d['private'] = private
        changed = True

    # Check if split_all_dirs
    split_all_dirs = i.get('split_all_dirs', '')
    if split_all_dirs != '':
        d['split_all_dirs'] = split_all_dirs
        changed = True

    # Check if recache
    xrecache = i.get('recache', '')
    if xrecache != '':
        d['recache'] = xrecache
        changed = True

    # Refreshing local repository description
    if remote != 'yes':
        pcfg = os.path.join(p, ck.cfg['repo_file'])
        if os.path.isfile(pcfg):
            if o == 'con':
                ck.out('')
                ck.out('Refreshing repo info in '+pcfg+' ...')

            r = ck.load_json_file({'json_file': pcfg})
            if r['return'] > 0:
                return r

            dcfg = r['dict']

            if 'dict' not in dcfg:
                dcfg['dict'] = {}

            dcfg['dict'].update(d)

            # Clean not needed keys (local)
            for k in ['data_uid', 'data_uoa', 'data_alias', 'data_name', 'path']:
                if k in dcfg['dict']:
                    del(dcfg['dict'][k])

            r = ck.save_json_to_file(
                {'json_file': pcfg, 'dict': dcfg, 'sort_keys': 'yes'})
            if r['return'] > 0:
                return ry

            # Update back repo meta in CK
            copy_dcfg = copy.deepcopy(dcfg)
            d.update(copy_dcfg['dict'])

    # Write if changed
#    if changed or i.get('update','')=='yes':
    if o == 'con':
        ck.out('')
        ck.out('Refreshing repo info in the CK local:repo:'+duoa+' ...')

    rx = ck.access({'action': 'update',
                    'module_uoa': ck.cfg['repo_name'],
                    'data_uoa': duoa,
                    'data_name': dn,
                    'dict': d,
                    'common_func': 'yes',
                    'substitute': 'yes',
                    'skip_update': 'yes',
                    'sort_keys': 'yes'})
    if rx['return'] > 0:
        return rx

    # Recaching
    if o == 'con':
        ck.out('')
        ck.out('Recaching repos to speed up access ...')
        ck.out('')
    r = recache({'out': o})
    if r['return'] > 0:
        return r

    return {'return': 0}

##############################################################################
# Pull from shared repo if URL


def pull(i):
    """
    Input:  {
              (path)  - repo UOA (where to create entry)
              (type)  - type
              (url)   - URL

                or

              (data_uoa)      - repo UOA

              (clone)         - if 'yes', clone repo instead of update

              (current_repos) - if resolving dependencies on other repos, list of repos being updated (to avoid infinite recursion)

              (git)           - if 'yes', use git protocol instead of https

              (ignore_pull)   - useful just for switching to another branch

              (stable)        - take stable version (highly experimental)
              (version)       - checkout version (default - stable)
              (branch)        - git branch
              (checkout)      - git checkout
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    xrecache = False

    pp = []
    px = i.get('path', '')
    t = i.get('type', '')
    url = i.get('url', '')

    stable = i.get('stable', '')
    version = i.get('version', '')
    if stable == 'yes':
        version = 'stable'
    branch = i.get('branch', '')
    checkout = i.get('checkout', '')

    ip = i.get('ignore_pull', '')

    cr = i.get('current_repos', [])

    tt = 'pull'
    if i.get('clone', '') == 'yes':
        tt = 'clone'

    if px != '':
        pp.append({'path': px, 'type': t, 'url': url})

    uoa = i.get('data_uoa', '')

    cids = i.get('cids', [])
    if len(cids) > 0 and uoa == '':
        uoa = cids[0]

    if ck.cfg.get('force_lower', '') == 'yes':
        uoa = uoa.lower()

    # If url is not empty and uoa is empty, get name from URL:
    if url != '' and uoa == '' and px == '':
        ix = url.rfind('/')
        if ix > 0:
            uoa = url[ix+1:]
            if uoa.endswith('.git'):
                uoa = uoa[:-4]

            i['data_uoa'] = uoa

    if uoa == '' and len(pp) == 0 and url == '':
        uoa = '*'

    if uoa != '':
        if uoa.find('*') >= 0 or uoa.find('?') >= 0:
            r = ck.list_data(
                {'module_uoa': work['self_module_uoa'], 'data_uoa': uoa})
            if r['return'] > 0:
                return r

            lst = r['lst']
            for q in lst:
                # Loading repo
                r = ck.access({'action': 'load',
                               'module_uoa': work['self_module_uoa'],
                               'data_uoa': q['data_uoa'],
                               'common': 'yes'})
                if r['return'] > 0:
                    return r
                d = r['dict']
                t = d.get('shared', '')
                duoa = r['data_uoa']

                if d.get('recache', '') == 'yes':
                    xrecache = True

                if t != '':
                    p = d.get('path', '')
                    url = d.get('url', '')
                    checkouts = d.get('checkouts', {})
                    pp.append({'path': p, 'type': t, 'url': url,
                               'data_uoa': duoa, 'checkouts': checkouts})
        else:
            # Loading repo
            r = ck.access({'action': 'load',
                           'module_uoa': work['self_module_uoa'],
                           'data_uoa': uoa,
                           'common': 'yes'})
            if r['return'] > 0:
                if r['return'] == 16:
                    # If not found, try to add from GIT

                    i['action'] = 'add'
                    i['shared'] = 'yes'
                    x = i.get('quiet', '')
                    if x == '':
                        x = 'yes'
                    i['quiet'] = x
                    i['current_repos'] = cr

                    return add(i)
                else:
                    return r

            d = r['dict']
            duoa = r['data_uoa']

            if d.get('recache', '') == 'yes':
                xrecache = True

            p = d['path']
            t = d.get('shared', '')
            url = d.get('url', '')
            checkouts = d.get('checkouts', {})

            pp.append({'path': p, 'type': t, 'url': url,
                       'data_uoa': duoa, 'checkouts': checkouts})

    # Updating ...
    for q in pp:
        p = q.get('path', '')
        duoa = q.get('data_uoa', '')
        t = q.get('type', '')
        url = q.get('url', '')

        # Substitute all https:// with git@ to be able to save password or use ssh key
        if i.get('git', '') == 'yes' or ck.cfg.get('use_git_instead_of_https', '') == 'yes':
            url = url.replace('https://', 'git@')

            j = url.find('/')
            if j > 0:
                url = url[:j]+':'+url[j+1:]

            url += '.git'

        # Check if repository is broken
        try:
            from ck import net
            r = net.request(
                {'get': {'action': 'get-repo-status', 'data_uoa': duoa, 'url': url}})
            if r['return'] == 0:
                s = r.get('dict', {}).get('warning', '')
                if s != '':
                    ck.out(
                        '******************************************************************')
                    ck.out('Warning: '+s)
        except Exception as e:
            pass

        if o == 'con' and tt != 'clone':
            ck.out('******************************************************************')
            ck.out('Updating repo "'+duoa+'" ...')
            ck.out('')
            ck.out('  Local path: '+p)
            ck.out('  URL:        '+url)

        if t == 'git':
            # Check if git is installed
            rq = ck.gen_tmp_file({})
            if rq['return'] > 0:
                return rq
            xfn = rq['file_name']

            os.system('git --version > '+xfn)

            rq = ck.load_text_file({'text_file': xfn,
                                    'delete_after_read': 'yes'})
            xs = ''
            if rq['return'] == 0:
                xs = rq['string'].strip()

            if xs.find(' version ') < 0:
                return{'return': 1, 'error': 'git command line client is not found - please, install it or download repo as zip'}

            # Continue
            try:
                px = os.getcwd()
            except OSError:
                from os.path import expanduser
                px = expanduser("~")

            if not os.path.isdir(p):
                os.makedirs(p)

            if o == 'con':
                ck.out('')
                ck.out('  cd '+p)
            os.chdir(p)

            r = 0
            if ip != 'yes':
                s = ck.cfg['repo_types'][t][tt].replace(
                    '$#url#$', url).replace('$#path#$', p)

                if o == 'con':
                    ck.out('  '+s)
                    ck.out('')

                r = os.system(s)

                if o == 'con':
                    ck.out('')

            os.chdir(px)  # Restore path

            if r > 0:
                if o == 'con':
                    ck.out('')
                    ck.out(
                        ' WARNING: repository update likely failed OR IN A DIFFERENT BRANCH/CHECKOUT (git exit code: '+str(r)+')')
                    ck.out('')
                    rx = ck.inp(
                        {'text': 'Would you like to continue (Y/n)?: '})
                    x = rx['string'].lower()
                    if x == 'n' or x == 'no':
                        return {'return': 1, 'error': 'repository update likely failed - exit code '+str(r)}
                else:
                    return {'return': 1, 'error': 'repository update likely failed - exit code '+str(r)}
        else:
            if o == 'con':
                ck.out('CK warning: this repository is not shared!')

        # Check deps
        if tt != 'clone':  # clone is done in add ...
            if o == 'con':
                ck.out('  ========================================')
                ck.out('  Checking dependencies on other repos ...')
                ck.out('')

            r = deps({'path': p,
                      'current_path': cr,
                      'how': 'pull',
                      'version': version,
                      'branch': branch,
                      'checkout': checkout,
                      'out': o})
            if r['return'] > 0:
                return r

    # Re-caching ...
    if xrecache:
        if o == 'con':
            ck.out('  ==============================================')
            ck.out('  At least one repository requires recaching ...')
            ck.out('')

        r = recache({'out': o})
        if r['return'] > 0:
            return r

    return {'return': 0}

##############################################################################
# Push and commit to shared repo if URL


def push(i):
    """
    Input:  {
              (path)  - repo UOA (where to create entry)
              (type)  - type
              (url)   - URL

                or

              (data_uoa)  - repo UOA

              (clone) - if 'yes', clone repo instead of update
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    pp = []
    px = i.get('path', '')
    t = i.get('type', '')
    url = i.get('url', '')

    if px != '':
        pp.append({'path': px, 'type': t, 'url': url})

    uoa = i.get('data_uoa', '')
    cids = i.get('cids', [])
    if len(cids) > 0:
        uoa = cids[0]

    if uoa == '' and len(pp) == 0:
        uoa = '*'

    if ck.cfg.get('force_lower', '') == 'yes':
        uoa = uoa.lower()

    if uoa != '':
        if uoa.find('*') >= 0 or uoa.find('?') >= 0:
            r = ck.list_data(
                {'module_uoa': work['self_module_uoa'], 'data_uoa': uoa})
            if r['return'] > 0:
                return r

            lst = r['lst']
            for q in lst:
                # Loading repo
                r = ck.access({'action': 'load',
                               'module_uoa': work['self_module_uoa'],
                               'data_uoa': q['data_uoa'],
                               'common': 'yes'})
                if r['return'] > 0:
                    return r
                d = r['dict']
                t = d.get('shared', '')
                if t != '':
                    p = d.get('path', '')
                    url = d.get('url', '')
                    pp.append({'path': p, 'type': t, 'url': url})
        else:
            # Loading repo
            r = ck.access({'action': 'load',
                           'module_uoa': work['self_module_uoa'],
                           'data_uoa': uoa,
                           'common': 'yes'})
            if r['return'] > 0:
                return r
            d = r['dict']

            p = d['path']
            t = d.get('shared', '')
            url = d.get('url', '')

            pp.append({'path': p, 'type': t, 'url': url})

    # Pushing ...
    for q in pp:
        p = q.get('path', '')
        t = q.get('type', '')
        url = q.get('url', '')

        if o == 'con':
            ck.out('')
            ck.out('Trying to commit and push '+p+' ...')

        if t == 'git':
            px = os.getcwd()

            if not os.path.isdir(p):
                return {'return': 1, 'error': 'local path to repository is not found'}

            if o == 'con':
                ck.out('')
                ck.out('cd '+p+' ...')

            os.chdir(p)

            s = ck.cfg['repo_types'][t]['commit'].replace(
                '$#url#$', url).replace('$#path#$', p)
            if o == 'con':
                ck.out('')
                ck.out('Executing command: '+s)
                ck.out('')
            r = os.system(s)

            if o == 'con':
                ck.out('')

            s = ck.cfg['repo_types'][t]['push'].replace(
                '$#url#$', url).replace('$#path#$', p)
            if o == 'con':
                ck.out('')
                ck.out('Executing command: '+s)
                ck.out('')
            r = os.system(s)

            if o == 'con':
                ck.out('')

            os.chdir(px)  # Restore path

            if r > 0:
                if o == 'con':
                    ck.out('')
                    ck.out(
                        ' WARNING: repository update likely failed - exit code '+str(r))
                    ck.out('')
                    rx = ck.inp(
                        {'text': 'Would you like to continue (Y/n)?: '})
                    x = rx['string'].lower()
                    if x == 'n' or x == 'no':
                        return {'return': 1, 'error': 'repository update likely failed - exit code '+str(r)}
                else:
                    return {'return': 1, 'error': 'repository update likely failed - exit code '+str(r)}
        else:
            if o == 'con':
                ck.out('CK warning: this repository is not shared!')

    return {'return': 0}

##############################################################################
# Create repository in a given directory and record info in CK


def create(i):
    """
    See function 'add'

    """

    return add(i)

##############################################################################
# Recache all repositories in cache


def recache(i):
    """
    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }
    """

    o = i.get('out', '')

    # Listing all repos
    r = ck.access({'action': 'list',
                   'module_uoa': ck.cfg['repo_name']})
    if r['return'] > 0:
        return r
    l = r['lst']

    cru = {}
    cri = {}

    # Processing repos
    # We need 2 passes (if some repos such as remote ones are referenced inside new repos)

    for ps in [0, 1]:
        for q in l:
            if ps == 0 or (ps == 1 and q.get('processed', '') != 'yes'):
                ruoa = q['repo_uoa']
                muoa = q['module_uoa']
                duoa = q['data_uoa']
                duid = q['data_uid']

                # First try to load from cache to check that not remote ...
                remote = False
                rx = ck.load_repo_info_from_cache({'repo_uoa': duoa})
                if rx['return'] == 0:
                    rd = rx.get('dict', {})
                    if rd.get('remote', '') == 'yes':
                        remote = True

                if not remote:
                    if duid == ck.cfg['repo_uid_default'] or duid == ck.cfg['repo_uid_local']:
                        if o == 'con':
                            ck.out('Skipping repo '+duoa+' ...')
                    else:
                        if o == 'con':
                            ck.out('Processing repo '+duoa+' ...')

                        # Repo dictionary (may be changing in .ckr.json)
                        dt = {}

                        # Find real repo and get .ckr.json
                        rx = ck.access({'action': 'where',
                                        'module_uoa': muoa,
                                        'data_uoa': duoa})
                        if rx['return'] == 0:
                            pckr = os.path.join(
                                rx['path'], ck.cfg['repo_file'])
                            if os.path.isfile(pckr):
                                rx = ck.load_json_file({'json_file': pckr})
                                if rx['return'] > 0:
                                    return rx

                                dt = rx['dict']['dict']

                        # Load extra info repo (do not use repo, since may not exist in cache)
                        rx = ck.access({'action': 'load',
                                        'module_uoa': muoa,
                                        'data_uoa': duoa})
                        if rx['return'] > 0:
                            if ps == 0:
                                continue
                            else:
                                return rx

                        if len(dt) == 0:
                            dt = rx['dict']
                        else:
                            if rx['dict'].get('path', '') != '':
                                dt['path'] = rx['dict']['path']

                        dname = rx['data_name']
                        dalias = rx['data_alias']
                        dp = rx['path']

                        if duoa != duid:
                            cru[duoa] = duid

                        dd = {'dict': dt}

                        dd['data_uid'] = duid
                        dd['data_uoa'] = duoa
                        dd['data_alias'] = dalias
                        dd['data_name'] = dname
                        dd['path_to_repo_desc'] = dp

                        cri[duid] = dd

                    q['processed'] = 'yes'

        # Recording
        ck.cache_repo_uoa = cru
        ck.cache_repo_info = cri

        rx = ck.save_repo_cache({})
        if rx['return'] > 0:
            return rx

        rx = ck.reload_repo_cache({'force': 'yes'})
        if rx['return'] > 0:
            return rx

    if o == 'con':
        ck.out('')
        ck.out('Repositories were successfully recached!')

    return {'return': 0}

##############################################################################
# Remove information about repository


def rm(i):
    """
    Input:  {
              (repo_uoa)            - repo UOA (where to delete entry about repository)
              uoa                   - data UOA
              (force)               - if 'yes', force removal
              (with_files) or (all) - if 'yes', remove files as well
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }
    """

    # Check if global writing is allowed
    r = ck.check_writing({})
    if r['return'] > 0:
        return r

    global cache_repo_uoa, cache_repo_info

    ruoa = i.get('repo_uoa', '')
    uoa = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        ruoa = ruoa.lower()
        uoa = uoa.lower()

    o = i.get('out', '')

    if uoa == '':
        return {'return': 1, 'error': 'UOA of the repository is not defined'}

    wf = i.get('with_files', '')
    if wf == '':
        wf = i.get('all', '')

    force = i.get('force', '')

    r = ck.access({'action': 'load',
                   'repo_uoa': ruoa,
                   'module_uoa': work['self_module_uoa'],
                   'data_uoa': uoa,
                   'common_func': 'yes'})
    if r['return'] > 0:
        return r
    duid = r['data_uid']
    duoa = r['data_uoa']

    d = r['dict']
    p = d.get('path', '')

    to_delete = True
    if o == 'con' and force != 'yes':
        r = ck.inp(
            {'text': 'Are you sure to delete information about repository '+duoa+' (y/N): '})
        c = r['string'].lower()
        if c != 'yes' and c != 'y':
            to_delete = False

    if to_delete and o == 'con' and force != 'yes' and wf == 'yes':
        r = ck.inp(
            {'text': 'You indicated that you want to DELETE ALL ENTRIES IN THE REPOSITORY! Are you sure (y/N): '})
        x = r['string'].lower()
        if x != 'yes' and x != 'y':
            wf = ''

    if to_delete:
        if o == 'con':
            ck.out('')
            ck.out('Reloading repo cache ...')
        r = ck.reload_repo_cache({})  # Ignore errors
        if r['return'] > 0:
            return r

        if o == 'con':
            ck.out('Removing from cache ...')
        if duoa in ck.cache_repo_uoa:
            del (ck.cache_repo_uoa[duoa])
        if duid in ck.cache_repo_info:
            del (ck.cache_repo_info[duid])

        if o == 'con':
            ck.out('Rewriting repo cache ...')
        r = ck.save_repo_cache({})
        if r['return'] > 0:
            return r

        if o == 'con':
            ck.out('Removing entry ...')
        r = ck.access({'action': 'remove',
                       'repo_uoa': ruoa,
                       'module_uoa': work['self_module_uoa'],
                       'data_uoa': uoa,
                       'common_func': 'yes'})
        if r['return'] > 0:
            return r

        if wf == 'yes' and p != '':
            if o == 'con':
                ck.out('Removing entries from the repository ...')
            import shutil
            if os.path.isdir(p):
                shutil.rmtree(p, onerror=ck.rm_read_only)

        if o == 'con':
            ck.out('')
            ck.out('Information about repository was removed successfully!')
            if wf != 'yes':
                ck.out('Note: repository itself was not removed!')

    return {'return': 0}

##############################################################################
# Remove information about repository


def remove(i):
    """
    Input:  { See 'rm' function }
    Output: { See 'rm' function }
    """

    return rm(i)

##############################################################################
# Remove information about repository


def delete(i):
    """
    Input:  { See 'rm' function }
    Output: { See 'rm' function }
    """

    return rm(i)

##############################################################################
# find path to a local repository


def where(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    r = ck.find_path_to_repo({'repo_uoa': duoa})
    if r['return'] > 0:
        return r

    d = r.get('dict', {})
    url = d.get('url', '')

    p = r['path']
    if p == '':
        p = url

    if o == 'con':
        ck.out(p)

    return r

##############################################################################
# archive repository


def zip(i):
    """
    Input:  {
              data_uoa       - repo UOA

              (archive_path) - if '' create inside repo path

              (archive_name) - if !='' use it for zip name
              (auto_name)    - if 'yes', generate name name from data_uoa: ckr-<repo_uoa>.zip
              (bittorent)    - if 'yes', generate zip name for BitTorrent: ckr-<repo_uid>-YYYYMMDD.zip

              (overwrite)    - if 'yes', overwrite zip file
              (store)        - if 'yes', store files instead of packing


              (data)         - CID allowing to add only these entries with pattern (can be from another archive)

              (all)          - archive all files
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    duoa1 = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa1 = duoa1.lower()

    duid = ''
    path = ''

    if duoa1 != '':
        # Find path to repo
        r = ck.find_path_to_repo({'repo_uoa': duoa1})
        if r['return'] > 0:
            return r

        duoa = r['repo_uoa']
        duid = r['repo_uid']
        path = r['path']

    an = i.get('archive_name', '')
#    if an=='': an='ckr.zip'

    if i.get('auto_name', '') == 'yes':
        an = 'ckr'
        if duoa != '':
            an += '-'+duoa
        an += '.zip'
    elif i.get('bittorrent', '') == 'yes':
        import time
        an = 'ckr-'
        if duid != '':
            an += duid
        an += '-'+time.strftime('%Y%m%d')+'.zip'
    elif an == '':
        if duoa1 == '':
            an = 'ckr.zip'
        else:
            an = 'ckr-'+duoa1+'.zip'

    ap = i.get('archive_path', '')
#    if ap=='': ap=path

    pfn = os.path.join(ap, an)

    if pfn.startswith('~'):
        from os.path import expanduser
        home = expanduser("~")

        pfn = os.path.abspath(home+os.sep+pfn[1:])

    if os.path.isfile(pfn):
        if i.get('overwrite', '') == 'yes':
            os.remove(pfn)
        else:
            return {'return': 1, 'error': 'archive '+pfn+' already exists'}

    if o == 'con':
        ck.out('Creating archive '+pfn +
               ' - please wait, it may take some time ...')

    # Check all files
    ignore = []
    if i.get('all', '') != 'yes':
        ignore = ck.cfg.get('ignore_directories_when_archive_repo', [])

    # Prepare archive
    import zipfile

    zip_method = zipfile.ZIP_DEFLATED
    if i.get('store', '') == 'yes':
        zip_method = zipfile.ZIP_STORED

    # Prepare list of files
    fl = {}

    data = i.get('data', '')
    if data != '':
        xpm = {}

        rx = ck.access({'action': 'search',
                        'cid': data})
        if rx['return'] > 0:
            return rx
        lst = rx['lst']
        for q in lst:
            pp = q['path']

            pm1, pd = os.path.split(pp)
            pr, pm = os.path.split(pm1)

            if pr not in fl:
                fl[pr] = []

            ry = ck.find_path_to_entry({'path': pr, 'data_uoa': pm})
            if ry['return'] > 0:
                return ry
            pm_uid = ry['data_uid']
            pm_alias = ry['data_alias']

            if pm_alias != '':
                if pm_alias not in xpm:
                    xpm[pm_alias] = pm_uid
                    fl[pr].append(os.path.join(
                        ck.cfg['subdir_ck_ext'], ck.cfg['file_alias_a'] + pm_alias))
                    fl[pr].append(os.path.join(
                        ck.cfg['subdir_ck_ext'], ck.cfg['file_alias_u'] + pm_uid))

            ry = ck.find_path_to_entry({'path': pm1, 'data_uoa': pd})
            if ry['return'] > 0:
                return ry
            pd_uid = ry['data_uid']
            pd_alias = ry['data_alias']

            if pd_alias != '':
                fl[pr].append(os.path.join(
                    pm, ck.cfg['subdir_ck_ext'], ck.cfg['file_alias_a'] + pd_alias))
                fl[pr].append(os.path.join(
                    pm, ck.cfg['subdir_ck_ext'], ck.cfg['file_alias_u'] + pd_uid))

            r = ck.list_all_files(
                {'path': pp, 'all': 'yes', 'ignore_names': ignore})
            if r['return'] > 0:
                return r
            for q in r['list']:
                fx = os.path.join(pm, pd, q)
                fl[pr].append(fx)

    else:
        r = ck.list_all_files(
            {'path': path, 'all': 'yes', 'ignore_names': ignore})
        if r['return'] > 0:
            return r
        fl[path] = r['list']

    # Write archive
    try:
        f = open(pfn, 'wb')
        z = zipfile.ZipFile(f, 'w', zip_method)
        for path in fl:
            fl1 = fl[path]
            for fn in fl1:
                p1 = os.path.join(path, fn)
                z.write(p1, fn, zip_method)
        z.close()
        f.close()

    except Exception as e:
        return {'return': 1, 'error': 'failed to prepare archive ('+format(e)+')'}

    return {'return': 0}

##############################################################################
# unzip entries to a given repo


def unzip(i):
    """
    Input:  {
              (data_uoa)    - repo UOA where to unzip (default, if not specified)
              zip           - path to zipfile (local or remote http/ftp)
              (overwrite)   - if 'yes', overwrite files when unarchiving
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')
    if duoa == '':
        duoa = 'local'

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    overwrite = i.get('overwrite', '')

    zip_ = i.get('zip', '')
    if zip_ == '':
        zip_ = 'ckr.zip'

    # Check if exists
    if not os.path.isfile(zip_):
       return {'return':1, 'error':'file "'+zip_+'" not found'}

    # Find path to repo
    r = ck.find_path_to_repo({'repo_uoa': duoa})
    if r['return'] > 0:
        return r

    path = r['path']

    # Unzipping archive
    rz = get_and_unzip_archive(
        {'zip': zip_, 'path': path, 'overwrite': overwrite, 'out': o})
    if rz['return'] > 0:
        return rz

    return {'return': 0}

##############################################################################
# received from web (if needed) and unzip archive


def get_and_unzip_archive(i):
    """
    Input:  {
              zip              - zip filename or URL
              path             - path to extract
              (overwrite)      - if 'yes', overwrite files when unarchiving
              (path_to_remove) - if !='', remove this part of the path from extracted archive
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    zp = i['zip']
    p = i['path']
    pr = i.get('path_to_remove', '')

    overwrite = i.get('overwrite', '')

    # If zip, get (download) and unzip file ...
    rm_zip = False
    if zp.find('://') >= 0:
        if o == 'con':
            ck.out('Downloading CK archive ('+zp+') - it may take some time ...')

        rm_zip = True

        # Generate tmp file
        import tempfile
        # suffix is important - CK will delete such file!
        fd, fn = tempfile.mkstemp(suffix='.tmp', prefix='ck-')
        os.close(fd)
        os.remove(fn)

        # Import modules compatible with Python 2.x and 3.x
        import urllib

        try:
            import urllib.request as urllib2
        except:
            import urllib2

        # Prepare request
        request = urllib2.Request(zp)

        # Connect
        try:
            f = urllib2.urlopen(request)
        except Exception as e:
            return {'return': 1, 'error': 'Failed downloading CK archive ('+format(e)+')'}

        import time
        t = time.time()
        t0 = t

        chunk = 32767
        size = 0

        try:
            fo = open(fn, 'wb')
        except Exception as e:
            return {'return': 1, 'error': 'problem opening file='+fn+' ('+format(e)+')'}

        # Read from Internet
        try:
            while True:
                s = f.read(chunk)
                if not s:
                    break
                fo.write(s)

                size += len(s)

                if o == 'con' and (time.time()-t) > 3:
                    speed = '%.1d' % (size/(1000*(time.time()-t0)))
                    ck.out('  Downloaded '+str(int(size/1000)) +
                           ' KB ('+speed+' KB/sec.) ...')
                    t = time.time()

            f.close()
        except Exception as e:
            return {'return': 1, 'error': 'Failed downlading CK archive ('+format(e)+')'}

        fo.close()

        zp = fn

    # Unzip if zip
    if zp != '':
        if o == 'con':
            ck.out('  Extracting to '+p+' ...')

        import zipfile
        f = open(zp, 'rb')
        z = zipfile.ZipFile(f)

        # First, try to find .ckr.json
        xprefix = ''
        for dx in z.namelist():
            if pr != '' and dx.startswith(pr):
                dx = dx[len(pr):]
            if dx.endswith(ck.cfg['repo_file']):
                xprefix = dx[:-len(ck.cfg['repo_file'])]
                break

        # Second, extract files
        for dx in z.namelist():
            dx1 = dx
            if pr != '' and dx1.startswith(pr):
                dx1 = dx1[len(pr):]
            if xprefix != '' and dx1.startswith(xprefix):
                dx1 = dx1[len(xprefix):]

            if dx1 != '':
                pp = os.path.join(p, dx1)
                if dx.endswith('/'):
                    # create directory
                    if not os.path.exists(pp):
                        os.makedirs(pp)
                else:
                    # extract file
                    ppd = os.path.dirname(pp)
                    if not os.path.exists(ppd):
                        os.makedirs(ppd)

                    if os.path.isfile(pp) and overwrite != 'yes':
                        if o == 'con':
                            ck.out(
                                'File '+dx+' already exists in the entry - skipping ...')
                    else:
                        fo = open(pp, 'wb')
                        fo.write(z.read(dx))
                        fo.close()
        f.close()

        if rm_zip:
            os.remove(zp)

    return {'return': 0}

##############################################################################
# import repo from current path


def import_repo(i):
    """
    Input:  {
              See action 'add' where import=yes
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['import'] = 'yes'
    return add(i)

##############################################################################
# resolve dependencies for a repo


def deps(i):
    """
    Input:  {
              (data_uoa)      - repo UOA
                  or
              (path)          - path to .cmr.json

              (current_repos) - list of repos being updated (to avoid infinite recursion)

              (how)           - 'pull' (default) or 'add'

              (version)       - checkout version (default - stable)
              (branch)        - git branch
              (checkout)      - git checkout
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    # Added repos to avoid duplication/recursion
    cr = i.get('current_repos', [])

    how = i.get('how', '')
    if how == '':
        how = 'pull'

    p = i.get('path', '')
    if p == '':
        r = ck.access({'action': 'load',
                       'module_uoa': work['self_module_uoa'],
                       'data_uoa': duoa})
        if r['return'] > 0:
            return r
        dr = r['dict']
        p = dr.get('path', '')

    if p != '':
        # path to repo description
        pp = os.path.join(p, ck.cfg['repo_file'])
        if os.path.isfile(pp):
            r = ck.load_json_file({'json_file': pp})
            if r['return'] > 0:
                return r

            d = r['dict']

            # Check checkouts
            version = i.get('version', '')
            branch = i.get('branch', '')
            checkout = i.get('checkout', '')

            if version != '':
                cx = d.get('dict', {}).get('checkouts', {}).get(version, {})
                branch = cx.get('branch', '')
                checkout = cx.get('checkout', '')

            ppp = os.getcwd()

            os.chdir(p)

            if branch != '':
                if o == 'con':
                    ck.out('  ====================================')
                    ck.out('  git checkout '+branch)
                    ck.out('')

                r = ck.run_and_get_stdout({'cmd': ['git', 'checkout', branch]})
                ck.out(r.get('stdout', ''))
                ck.out(r.get('stderr', ''))

            if checkout != '':
                if o == 'con':
                    ck.out('  ====================================')
                    ck.out('  git checkout '+checkout)
                    ck.out('')

                r = ck.run_and_get_stdout(
                    {'cmd': ['git', 'checkout', checkout]})
                ck.out(r.get('stdout', ''))
                ck.out(r.get('stderr', ''))

            os.chdir(ppp)

            rp1 = d.get('dict', {}).get('repo_deps', [])
            if len(rp1) == 0:
                rp1 = d.get('repo_deps', [])  # for backwards compatibility ...

            rp2 = []
            rp = []

            if len(rp1) > 0:
                for xruoa in rp1:
                    if type(xruoa) != list:  # for backwards compatibility
                        ruoa = xruoa.get('repo_uoa', '')
                        if xruoa.get('repo_uid', '') != '':
                            ruoa = xruoa['repo_uid']
                        if ruoa != '' and ruoa not in cr:
                            rp2.append(xruoa)

            # Add dependencies on other repositories (but avoid duplication)
            if len(rp2) == 0:
                if o == 'con':
                    ck.out('  No dependencies on other repositories found!')
            else:
                for xruoa in rp2:
                    ruoa = xruoa.get('repo_uoa', '')
                    if xruoa.get('repo_uid', '') != '':
                        ruoa = xruoa['repo_uid']
                    rurl = xruoa.get('repo_url', '')
                    if ruoa != '':
                        x = '  Dependency on repository '+ruoa+' '

                        # Check if this repo exists
                        r = ck.access({'action': 'load',
                                       'module_uoa': work['self_module_uoa'],
                                       'data_uoa': ruoa})
                        if r['return'] > 0:
                            if r['return'] != 16:
                                return r
                            rp.append(xruoa)
                            x += ': should be resolved ...'
                        else:
                            # If explicit branch, still add !
                            branch = xruoa.get('branch', '')
                            checkout = xruoa.get('checkout', '')
                            stable = xruoa.get('stable', '')
                            version = xruoa.get('version', '')

                            if branch != '' or checkout != '' or stable != '' or version != '':
                                xruoa['ignore_pull'] = 'yes'
                                rp.append(xruoa)
                                x += ': should be switched to explicit branch ...'
                            else:
                                x += ': Ok'

                        if o == 'con':
                            ck.out(x)

            if len(rp) > 0:
                for xruoa in rp:
                    ruoa = xruoa.get('repo_uoa', '')
                    ruid = xruoa.get('repo_uid', '')
                    rurl = xruoa.get('repo_url', '')

                    branch = xruoa.get('branch', '')
                    checkout = xruoa.get('checkout', '')
                    stable = xruoa.get('stable', '')
                    version = xruoa.get('version', '')

                    ignore_pull = xruoa.get('ignore_pull', '')

                    if o == 'con':
                        ck.out('')
                        x = ''
                        if ruid != '':
                            x = ' ('+ruid+')'
                        ck.out('  Resolving dependency on repo: '+ruoa+x)
                        ck.out('')

                    if ruid != '':
                        cr.append(ruid)
                    else:
                        cr.append(ruoa)

                    ii = {'action': how,
                          'module_uoa': work['self_module_uoa'],
                          'data_uoa': ruoa,
                          'current_repos': cr,
                          'url': rurl,
                          'ignore_pull': ignore_pull,
                          'branch': branch,
                          'checkout': checkout,
                          'stable': stable,
                          'version': version,
                          'out': o}
                    if ruid != '':
                        ii['data_uid'] = ruid
                    if how == 'add':
                        ii['gitzip'] = 'yes'
                    r = ck.access(ii)
                    if r['return'] > 0:
                        return r

    return {'return': 0, 'current_repos': cr}

##############################################################################
# print dependencies on other shared repositories


def print_deps(i):
    """
    Input:  {
              data_uoa                   - data UOA of the repo
                 or
              repo_deps                  - dict with dependencies on other shared repos

              (out_prefix)               - output prefix befor each string
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              repo_deps
            }

    """

    o = i.get('out', '')

    op = i.get('out_prefix', '')

    duoa = i.get('data_uoa', '')
    if duoa != '':
        # Get configuration
        r = ck.load_repo_info_from_cache({'repo_uoa': duoa})
        if r['return'] > 0:
            return r

        d = r['dict']
        rp1 = d.get('dict', {}).get('repo_deps', [])
    else:
        rp1 = i['repo_deps']

    if len(rp1) == 0:
        rp1 = d.get('repo_deps', [])  # for compatibility ...

    if o == 'con' and len(rp1) > 0:
        for q in rp1:
            ruoa = q.get('repo_uoa', '')
            ruid = q.get('repo_uid', '')
            rurl = q.get('repo_url', '')

            x = op+ruoa
            if ruid != '':
                x += '; '+ruid
            elif rurl != '':
                x += '; '
            if rurl != '':
                x += '; '+rurl

            ck.out(x)

    return {'return': 0, 'repo_deps': rp1}

##############################################################################
# add more dependencies


def add_more_deps(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              repo_deps    - list with dependencies on other repositories ...
            }

    """

    rp = []

    r = ck.inp(
        {'text': 'Would you like to add extra dependencies on other shared repositories (y/N)?: '})
    x = r['string'].lower()
    if x == 'yes' or x == 'y':
        ck.out('')
        ck.out('Use the following format: repo UOA; (repo UID) (; repo URL)')
        ck.out('For example:')
        ck.out('  ck-autotuning')
        ck.out(
            '  ck-dissemination-modules;;https://github.com/gfursin/ck-dissemination-modules.git')
        ck.out('')
        ck.out('Press Enter to stop adding repositories!')
        ck.out('')

        while True:
            r = ck.inp({'text': ''})
            x = r['string'].strip()
            if x == '':
                break

            z = {}

            y = x.split(';')
            if len(y) > 0:
                z['repo_uoa'] = y[0].strip()
                if len(y) > 1:
                    z['repo_uid'] = y[1].strip()
                    if len(y) > 2:
                        z['repo_url'] = y[2].strip()

            if len(z) > 0:
                rp.append(z)

    return {'return': 0, 'repo_deps': rp}

##############################################################################
# Find location of the repository (see ticket #45)
# Technically speaking it doesn't follow CK concept
# to always find path to CK entries (and thus to the repo description),
# but it is more intuitive.
# That's why we also added option --meta to revert back
# to original functionality and find path to repo meta


def find(i):
    """
    Input:  {
              (meta)       - if 'yes', return location of repo entry with meta
                             rather than path to the repository
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    if i.get('meta', '') != 'yes':
        i['action'] = 'where'
    else:
        i['common_func'] = 'yes'

    return ck.access(i)

##############################################################################
# renew repository (remove fully and pull again)


def renew(i):
    """
    Input:  {
              data_uoa                   - data UOA of the repo

              (stable)                   - take stable version (highly experimental)
              (checkout)                 - checkout (default - stable)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # Check if global writing is allowed
    r = ck.check_writing({})
    if r['return'] > 0:
        return r

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')
    if duoa == '':
        return {'return': 1, 'error': 'repository UOA is not specified'}

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    # Get configuration (not from Cache - can be outdated info!)
#    r=ck.load_repo_info_from_cache({'repo_uoa':duoa})
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uoa'],
                   'data_uoa': duoa})
    if r['return'] > 0:
        return r

    p = r.get('dict', {}).get('path', '')
    d = r['dict']
    dn = r.get('data_name', '')

    shared = d.get('shared', '')
    url = d.get('url', '')

    if shared != 'git' and url == '':
        return {'return': 1, 'error': 'this repository is not shared and can not be renewed'}

    # first delete
    ii = {'action': 'rm',
          'module_uoa': work['self_module_uoa'],
          'data_uoa': duoa,
          'all': 'yes'}
    if o == 'con':
        ii['out'] = 'con'
    r = ck.access(ii)
    if r['return'] > 0:
        return r

    # pull again
    ii = {'action': 'pull',
          'module_uoa': work['self_module_uoa'],
          'data_uoa': duoa,
          'data_name': dn,
          'url': url,
          'stable': i.get('stable', ''),
          'checkout': i.get('checkout', '')}
    if o == 'con':
        ii['out'] = 'con'
    return ck.access(ii)

##############################################################################
# show remote repo in a browser


def browse(i):
    """
    Input:  {
              data_uoa                   - data UOA of the repo
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')
    if duoa == '':
        return {'return': 1, 'error': 'repository UOA is not specified'}

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    # Get configuration (not from Cache - can be outdated info!)
#    r=ck.load_repo_info_from_cache({'repo_uoa':duoa})
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uoa'],
                   'data_uoa': duoa})
    if r['return'] > 0:
        return r

    p = r.get('dict', {}).get('path', '')
    d = r['dict']
    dn = r.get('data_name', '')

    shared = d.get('shared', '')
    url = d.get('url', '')

    if shared != 'git' and url == '':
        return {'return': 1, 'error': 'this repository is not shared'}

    import webbrowser
    webbrowser.open(url)

    return {'return': 0}

##############################################################################
# describe repository for Artifact Evaluation


def describe(i):
    """
    Input:  {
              (dict)  - dict with current repo description
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              (dict)  - updated dict with current repo description
            }

    """

    d = i.get('dict', {})

    a = d.get('artifact_pack_description', {})

    ck.out('')
    ck.out('Please, provide description of your artifact pack (see http://cTuning.org/ae for more details):')
    ck.out('')

    r = ck.inp(
        {'text': 'Enter title of your artifact pack (or associated paper): '})
    x = r['string']
    if x != '':
        a['title'] = x

    r = ck.inp({'text': 'List contributors: '})
    x = r['string']
    if x != '':
        a['contributors'] = x

    r = ck.inp({'text': 'Enter artifact abstract: '})
    x = r['string']
    if x != '':
        a['description'] = x

    r = ck.inp({'text': 'Enter license of the whole artifact pack: '})
    x = r['string']
    if x != '':
        a['license'] = x

    r = ck.inp({'text': 'Enter version/revision of the whole artifact pack: '})
    x = r['string']
    if x != '':
        a['version'] = x

    r = ck.inp({'text': 'How delivered (URL, DOI, OCRID, etc): '})
    x = r['string']
    if x != '':
        a['how_delivered'] = x

    r = ck.inp({'text': 'Describe software dependencies: '})
    x = r['string']
    if x != '':
        a['software_dependencies'] = x

    r = ck.inp({'text': 'Describe hardware dependencies: '})
    x = r['string']
    if x != '':
        a['hardware_dependencies'] = x

    r = ck.inp({'text': 'Describe data sets: '})
    x = r['string']
    if x != '':
        a['datasets'] = x

    r = ck.inp({'text': 'Describe installation procedure: '})
    x = r['string']
    if x != '':
        a['installation'] = x

    r = ck.inp({'text': 'Describe possible experiment parameterization: '})
    x = r['string']
    if x != '':
        a['parameterization'] = x

    r = ck.inp({'text': 'Describe experiment workflow: '})
    x = r['string']
    if x != '':
        a['experiment_workflow'] = x

    r = ck.inp({'text': 'Describe evaluation procedure and expected output: '})
    x = r['string']
    if x != '':
        a['evaluation'] = x

    r = ck.inp({'text': 'Acknowledgments: '})
    x = r['string']
    if x != '':
        a['acknowledgments'] = x

    r = ck.inp({'text': 'Misc notes: '})
    x = r['string']
    if x != '':
        a['notes'] = x

    d['artifact_pack_description'] = a

    return {'return': 0, 'dict': d}

##############################################################################
# form URL from hostname, hostext and port


def form_url(i):
    """
    Input:  {
              (url)                      - if type=='remote' or 'git', URL of remote repository or git repository
              (hostname)                 - if !='', automatically form url above (add http:// + /ck?)
              (port)                     - if !='', automatically add to url above
              (hostext)                  - if !='', add to the end of above URL instead of '/ck?' -
                                           useful when CK server is accessed via Apache2, IIS or other web servers
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              url          - formed URL
            }

    """

    url = i.get('url', '')

    if i.get('hostname', '') != '':
        url = 'http://'+i['hostname']

        if i.get('port', '') != '':
            url += ':'+i['port']

        if i.get('hostext', '') != '':
            url += '/'+i['hostext']
        else:
            url += '/ck?'

    return {'return': 0, 'url': url}

##############################################################################
# show repositories and their status


def show(i):
    """
    Input:  {
              (data_uoa) - repo UOA

              (reset)    - if 'yes', reset repos

              (stable)   - take stable version (highly experimental)
              (version)  - checkout version (default - stable)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o = i.get('out', '')

    curdir = os.getcwd()

    duoa = i.get('data_uoa', '')

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    reset = i.get('reset', '')

    stable = i.get('stable', '')
    version = i.get('version', '')
    if stable == 'yes':
        version = 'stable'

    r = ck.list_data({'module_uoa': work['self_module_uoa'],
                      'data_uoa': duoa})
    if r['return'] > 0:
        return r

    if o == 'con':
        ck.out('Please wait - it may take some time ...')
        ck.out('')

    r = ck.reload_repo_cache({})  # Ignore errors

    # Init header.
    pp = [({'branch': 'branch', 'origin': 'origin', 'checkout': 'local',
            'path': 'path', 'type': 'type', 'url': 'url', 'data_uoa': 'data_uoa'})]
    il = 0
    for q in ck.cache_repo_info:
        # Get repo info
        qq = ck.cache_repo_info[q]

        d = qq['dict']

        t = d.get('shared', '')

        if t != '':
            duoa = qq['data_uoa']

            if len(duoa) > il:
                il = len(duoa)

            p = d.get('path', '')
            url = d.get('url', '')

            branch = ''
            origin = ''
            checkout = ''

            if os.path.isdir(p):
                # Detect status
                os.chdir(p)

                if reset == 'yes':
                    r = ck.run_and_get_stdout(
                        {'cmd': ['git', 'checkout', 'master']})

                if version != '':
                    cx = qq.get('dict', {}).get(
                        'checkouts', {}).get(version, {})
                    branch = cx.get('branch', '')
                    checkout = cx.get('checkout', '')

                    if branch != '':
                        r = ck.run_and_get_stdout(
                            {'cmd': ['git', 'checkout', branch]})

                    if checkout != '':
                        r = ck.run_and_get_stdout(
                            {'cmd': ['git', 'checkout', checkout]})

                # FGG TBD: we may need to add explicit check for branch/checkout in repo_deps here?
                # OR MAYBE NOT - need to think ...

                # Get current branch
                r = ck.run_and_get_stdout(
                    {'cmd': ['git', 'rev-parse', '--abbrev-ref', 'HEAD']})
                if r['return'] == 0 and r['return_code'] == 0:
                    branch = r['stdout'].strip()

                # Get origin hash
                r = ck.run_and_get_stdout(
                    {'cmd': ['git', 'rev-parse', '--short', 'origin/HEAD']})
                if r['return'] == 0 and r['return_code'] == 0:
                    origin = r['stdout'].strip()

                # Get current hash (append '-dirty' on dirty working tree)
                r = ck.run_and_get_stdout(
                    {'cmd': ['git', 'describe', '--match=NeVeRmAtCh', '--always', '--abbrev', '--dirty']})
                if r['return'] == 0 and r['return_code'] == 0:
                    checkout = r['stdout'].strip()

            pp.append({'branch': branch, 'origin': origin, 'checkout': checkout,
                       'path': p, 'type': t, 'url': url, 'data_uoa': duoa})

    # Print
    for q in pp:
        name = q['data_uoa']

        x = name+' '*(il-len(name))

        branch = q.get('branch', '')
        origin = q.get('origin', '')
        checkout = q.get('checkout', '')
        url = q.get('url', '')

        if branch != '' or 'origin' or checkout != '' or url != '':
            x += ' ( '+branch+' ; '+origin+' ; '+checkout+' ; '+url+' )'

        ck.out(x)

    os.chdir(curdir)

    return {'return': 0}

##############################################################################
# reset git repos to default branch and latest checkout (or specific one)


def reset(i):
    """
    Input:  {
              (data_uoa) - repository UOA

              (stable)        - take stable version (highly experimental)
              (version)       - checkout version (default - stable)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['reset'] = 'yes'

    return show(i)

##############################################################################
# rename repo


def ren(i):
    """
    Input:  {
              data_uoa  - repo UOA

              (new_data_uoa)
                 or
              xcids[0]           - {'data_uoa'} - new data UOA
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    duoa = i.get('data_uoa', '')

    if duoa == '':
        return {'return': 1, 'error': 'repo is not defined'}

    if ck.cfg.get('force_lower', '') == 'yes':
        duoa = duoa.lower()

    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uid'],
                   'data_uoa': duoa})
    if r['return'] > 0:
        return r

    dd = r['dict']
    dp = r['path']
    duoa_real = r['data_uoa']
    dname = r['data_name']

    nduoa = i.get('new_data_uoa', '')

    if nduoa == '':
        xcids = i.get('xcids', [])
        if len(xcids) > 0:
            xcid = xcids[0]
            nduoa = xcid.get('data_uoa', '')

    if nduoa == '':
        xcids = i.get('cids', [])
        if len(xcids) > 0:
            nduoa = xcids[0]

    if nduoa == '':
        return {'return': 1, 'error': 'new repo name is not defined'}

    if ck.cfg.get('allowed_entry_names', '') != '':
        import re

        anames = ck.cfg.get('allowed_entry_names', '')

        if not re.match(nduoa, a):
            return {'return': 1, 'error': 'found disallowed characters in names (allowed: "'+anames+'")'}

    if ck.cfg.get('force_lower', '') == 'yes':
        nduoa = nduoa.lower()

    if nduoa == 'local' or nduoa == 'default':
        return {'return': 1, 'error': 'new repo name already exists'}

    # Check if such repo doesn't exist
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uid'],
                   'data_uoa': nduoa})
    if r['return'] == 0:
        return {'return': 1, 'error': 'repo already exists'}

    # Update .ckr.json
    dpp = dd.get('path', '')
    if dpp != '':
        pckr = os.path.join(dpp, ck.cfg['repo_file'])

        r = ck.load_json_file({'json_file': pckr})
        if r['return'] > 0:
            return r

        dckr = r['dict']

        x = dckr.get('data_uoa', '')
        if x != '' and x == duoa_real:
            dckr['data_uoa'] = nduoa

        x = dckr.get('data_alias', '')
        if x != '' and x == duoa_real:
            dckr['data_alias'] = nduoa

        x = dckr.get('data_name', '')
        if x != '' and x == duoa_real:
            dckr['data_name'] = nduoa

        r = ck.save_json_to_file({'json_file': pckr, 'dict': dckr})
        if r['return'] > 0:
            return r

    # Rename repo entry using internal command
    r = ck.access({'action': 'ren',
                   'module_uoa': work['self_module_uid'],
                   'data_uoa': duoa,
                   'new_data_uoa': nduoa,
                   'common_func': 'yes'})
    if r['return'] > 0:
        return r

    # Recache repos
    r1 = recache({'out': o})
    if r1['return'] > 0:
        return r1

    return r

##############################################################################
# rename repo


def rename(i):
    """
    Input:  {
               See "ck ren repo --help"
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return ren(i)

##############################################################################
# copy repo (forbidden)


def copy(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return cp(i)

##############################################################################
# copy repo (forbidden)


def cp(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return {'return': 1, 'error': 'copying CK repositories is forbidden to avoid duplication of CK entries. Create a new one instead and move CK entries there'}

##############################################################################
# init CK repo in a given directory


def new_init(i):
    """
    Input:  {
              (data_uoa) - repo name
              (data_uid) - repo UID
              (url)      - URL if shared
              (deps)     - list of deps on other repositories separated by comma
              (path)     - path where to initialize a repository
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # Check if global writing is allowed
    r = ck.check_writing({})
    if r['return'] > 0:
        return r

    # Check output
    o = i.get('out', '')
    oo = ''
    if o == 'con':
        oo = o

    # Check path
    path = os.getcwd()
    if i.get('path', '') != '':
        path = i['path']
        if not os.path.isdir(path):
            return {'return': 1, 'error': 'path not found'}

    # Check name
    data_uoa = i.get('data_uoa', '')
    data_uid = i.get('data_uid', '')

    if data_uoa == '':
        # Check from .ckr.json
        pckr = os.path.join(path, ck.cfg['repo_file'])
        if os.path.isfile(pckr):
            r = ck.load_json_file({'json_file': pckr})
            if r['return'] > 0:
                return r
            d = r['dict']

            data_uoa = d.get('data_uoa', '')
            data_uid = d.get('data_uid', '')

            if data_uoa != '':
                ck.out('Detected CK repo name: '+data_uoa)

    if data_uoa == '':
        # Try to detect from .ckr.json
        r = ck.inp({'text': 'Enter CK repository name: '})
        if r['return'] > 0:
            return r
        data_uoa = r['string'].strip()

    # Check if already exists
    exists = False
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uid'],
                   'data_uoa': data_uoa})
    if r['return'] > 0 and r['return'] != 16:
        return r
    if r['return'] == 0:
        exists = True

        r = ck.inp({'text': 'CK repository "'+data_uoa +
                    '" already exists. Update (y/N)? '})
        if r['return'] > 0:
            return r
        s = r['string'].strip().lower()

        if s == '' or s == 'n' or s == 'no':
            return {'return': 1, 'error': 'CK repository already exists'}

    # Check URL
    url = i.get('url', '')
    if url == '':
        # Attempt to check from current Git dir
        pgit = os.path.join(path, '.git', 'config')
        if os.path.isfile(pgit):
            r = ck.load_text_file({'text_file': pgit, 'split_to_list': 'yes'})
            if r['return'] > 0:
                return r
            ll = r['lst']

            for l in ll:
                l = l.strip()
                if l.startswith('url ='):
                    url = l[5:].strip()
                    ck.out('Detected URL: '+url)
                    break

    # prepare deps
    repo_deps = []

    deps = i.get('deps', '').strip().split(',')

    for q in deps:
        q = q.strip()
        if q != '':
            repo_deps.append({'repo_uoa': q})

    # Create or update repo
    ii = {'module_uoa': work['self_module_uid'],
          'data_uoa': data_uoa,
          'quiet': 'yes',
          'path': path,
          'out': oo}

    if data_uid != '':
        ii['data_uid'] = data_uid

    if exists:
        ii['action'] = 'update'
    else:
        ii['action'] = 'add'

    if url != '':
        ii['shared'] = 'yes'
        ii['url'] = url

    if len(repo_deps) > 0:
        ii['repo_deps'] = repo_deps

    r = ck.access(ii)
    return r



# -*- coding: utf-8 -*-

#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

import unittest

ck = None  # Will be updated by CK (initialized CK kernel)

# Contains original tests, that where in the 'ck/tests' directory (not all of them).
# Some tests are simplified, because the original tests where in fact integration tests, which can not be easily
# represented as unit tests.
# Some tests are adapted to the API changes happened since then.
# Do *not* add new tests here.
# Methods are named like this: 'test_' + the original file name without extension.


class OriginalKernelTest(unittest.TestCase):

    def test_flatten_dict(self):
        a = {'dyn_features': {'ft1': '1', 'ft2': '2'},
             'static_features': {'ft3': '3', 'ft4': '4'}}
        r = ck.flatten_dict({'dict': a})
        self.assertEqual(0, r['return'], r.get('error', None))

        x = r['dict']
        self.assertEqual({'##dyn_features#ft2': '2', '##static_features#ft3': '3',
                          '##static_features#ft4': '4', '##dyn_features#ft1': '1'}, x)

        r = ck.restore_flattened_dict({'dict': x})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(a, r['dict'])

        r = ck.flatten_dict({'dict': ['t1', 't2']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'#@0': 't1', '#@1': 't2'}, r['dict'])

        r = ck.flatten_dict({'dict': 't1', 'prune_keys': [
                            'prefix'], 'prefix': 'prefix'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'prefix': 't1'}, r['dict'])

        r = ck.flatten_dict({'dict': 't1', 'prune_keys': [
                            'pr??ix'], 'prefix': 'prefix'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'prefix': 't1'}, r['dict'])

        r = ck.get_by_flat_key({'dict': ['t1', ['a1', 'a2']], 'key': '@1@1'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('a2', r['value'])

        d = ['t1']
        r = ck.set_by_flat_key({'dict': d, 'key': '@1#a@1@1', 'value': 'a3'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('a3', d[1]['a'][1][1])

    def test_list_with_ranges(self):
        r = ck.list_data({"repo_uoa_list": ["default"], "module_uoa_list": [
                         "test", "index"], "data_uoa_list": ["unicode"]})
        self.assertEqual(0, r['return'], r.get('error', None))

        lst = r['lst']
        print(lst)
        self.assertEqual(1, len(lst))
        self.assertEqual('unicode', lst[0]['data_uoa'])

    def test_test(self):
        import os

        r = ck.get_version({})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(ck.__version__, r['version_str'])
        self.assertEqual(ck.__version__.split('.'), r['version'])

        fname = ck.gen_tmp_file(
            {'prefix': 'ck-test-', 'suffix': '.json'})['file_name']
        try:
            with open(fname, 'w') as f:
                f.write('{"key20":["value21","value22"],  "key12":"12"}')
            cmd = "mv data cid1 cid2 key1=value1 -key10 --key13=value13 @" + fname
            i = cmd.split(' ')
            r = ck.convert_ck_list_to_dict(i)
            self.assertEqual({'cid': 'data', 'key1': 'value1', u'key20': [u'value21', u'value22'], 'action': 'mv', 'cids': [
                             'cid1', 'cid2'], 'key13': 'value13', u'key12': u'12', 'key10': 'yes'}, r['ck_dict'])
        finally:
            os.remove(fname)



#
# Collective Knowledge (check conditions)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# check conditions

def check(i):
    """
    Input:  {
              original_points - dict with original points
              new_points      - dict with new points
              results         - results for all points (from experiment)
              conditions      - list of conditions
              (middle_key)    - add this to keys in conditions (min, exp, mean, etc)
            }

    Output: {
              return           - return code =  0, if successful
                                             >  0, if error
              (error)          - error text if return > 0

              good_points      - list of good points
              points_to_delete - list of new points to delete

              keys             - list of checked keys
            }

    """

    o=i.get('out','')

    points1=i.get('original_points',[])
    points2=i['new_points']
    cc=i['conditions']
    results=i['results']

    mk=i.get('middle_key','')
    if mk=='': mk='#min'

    new=False
    points=[]  # good points (with correct conditions)
    dpoints=[] # points to delete if do not match conditions

    # Check keys
    keys=[]
    for c in cc:
        if len(c)!=4:
           import json
           return {'return':1, 'error':'condition length !=4 ('+json.dumps(c)+')'}

        kt=(c[0]+c[1]).replace('$#objective#$',mk)
        if kt not in keys:
           keys.append(kt)

    if o=='con':
       ck.out('')

    # Check conditions
    for q in points2:
        if q not in points1:
           # Find point in results
           qq={}
           for k in results:
               if k.get('point_uid','')==q:
                  qq=k
                  break

           if len(qq)>0:
              fine=True
              # Go over conditions

              for c in cc:
                  kt=(c[0]+c[1]).replace('$#objective#$',mk)

                  if kt not in keys:
                     keys.append(kt)

                  x=c[2]
                  y=c[3]

                  behavior=qq.get('flat',{})

                  dv=behavior.get(kt,None)

                  s='         - Condition on "'+kt+' '+str(x)+' '+str(y)+'" : '

                  if dv==None:
                     fine=False

                  if fine and x=='<' and dv>=y:
                     fine=False

                  if fine and (x=='<=' or x=='=<') and dv>y:
                     fine=False

                  if fine and x=='==' and dv!=y:
                     fine=False

                  if fine and x=='!=' and dv==y:
                     fine=False

                  if fine and x=='>' and dv<=y:
                     fine=False

                  if fine and (x=='>=' or x=='=>') and dv<y:
                     fine=False

                  if o=='con':
                     if fine:
                        s+='ok'
                     else:
                        s+='FAILED'

                     s+=' ('+str(dv)+')'
                     ck.out(s)

                  if not fine:
                     break

              if fine:
                 points.append(q)
              else:
                 dpoints.append(q)

    return {'return':0, 'good_points':points, 'points_to_delete':dpoints, 'keys':keys}



#
# Collective Knowledge (platform - CPU)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# detect CPU

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (skip_print_os_info)   - if 'yes', do not print OS info

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (extra_info)           - extra info about author, etc (see add from CK kernel)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              features = {
                cpu        - CPU features (properties), unified
                cpu_misc   - assorted CPU features (properties), platform dependent

                os         - OS features (properties), unified
                os_misc    - assorted OS features (properties), platform dependent
              }
            }

    """

    import os

    o=i.get('out','')

    oo=''
    if o=='con': oo=o

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    device_cfg=i.get('device_cfg',{})

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    sic=i.get('skip_info_collection','')
    sdi=i.get('skip_device_init','')
    pdv=i.get('print_device_info','')

    ex=i.get('exchange','')
    if ex=='': ex=i.get('share','')

    einf=i.get('extra_info','')
    if einf=='': einf={}

    # Get OS info ##############################################################
    import copy
    ii=copy.deepcopy(i)
    ii['out']=oo
    if i.get('skip_print_os_info','')=='yes':ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.os']
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tosd.update(device_cfg.get('update_target_os_dict',{}))

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    prop=rr['features']['os']

    # Some params
    ro=tosd.get('redirect_stdout','')
    remote=tosd.get('remote','')
    remote_ssh=tosd.get('remote_ssh','')
    win=tosd.get('windows_base','')
    mac=tosd.get('macos','')
    unix=win!='yes' and mac!='yes'

    dv=''
    if tdid!='': dv=' -s '+tdid

    # Init
    target={}
    target_freq={}
    target_freq_all={}
    target_freq_max={}
    target_num_proc=''
    info_cpu={}

    new_format=''
    unique_cpus=[]

    spd=tosd.get('skip_platform_detection','')

    if spd=='yes':
       # Nothing
       pass
    elif unix:
       # Get all params
       params={}
       if remote=='yes' and remote_ssh!='yes':
          rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
          if rx['return']>0: return rx
          fn=rx['file_name']

          x=tosd.get('adb_all_params','')
          x=x.replace('$#redirect_stdout#$', ro)
          x=x.replace('$#output_file#$', fn)

          dv=''
          if tdid!='': dv=' -s '+tdid
          x=x.replace('$#device#$',dv)

          if o=='con' and pdv=='yes':
             ck.out('')
             ck.out('Receiving all parameters:')
             ck.out('  '+x)

          rx=os.system(x)
          if rx!=0:
             if o=='con':
                ck.out('')
                ck.out('Non-zero return code :'+str(rx)+' - likely failed')
             return {'return':1, 'error':'access to remote device failed'}

          # Read and parse file
          rx=ck.load_text_file({'text_file':fn, 'split_to_list':'yes', 'delete_after_read':'yes'})
          if rx['return']>0: return rx
          ll=rx['lst']

          for s in ll:
              s1=s.strip()

              q2=s1.find(']: [')
              k=''
              if q2>=0:
                 k=s1[1:q2].strip()
                 v=s1[q2+4:].strip()
                 v=v[:-1].strip()

                 params[k]=v

       # Read cpuinfo
       fnx='/proc/cpuinfo'

       if remote=='yes':

          # Read file
          rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
          if rx['return']>0: return rx
          fcpuinfo=rx['file_name']

          x=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat '+fnx+' '+tosd.get('remote_shell_end','')+' '+ro+fcpuinfo

          if o=='con' and pdv=='yes':
             ck.out('')
             ck.out('Executing: '+x)

          rx=os.system(x)
          if rx!=0: 
             if os.path.isfile(fnx): os.remove(fcpuinfo)
             fcpuinfo='' # Do not process further

       else:
          fcpuinfo=fnx

       # Read and parse file
       pp=0 # current logical processor
       spp=str(pp)
       info_cpu[spp]={}
       target_freq[spp]=0
       target_freq_max[spp]=0
       first_skipped=False

       if fcpuinfo!='':
          rx=ck.load_text_file({'text_file':fcpuinfo, 'split_to_list':'yes'})
          if rx['return']>0: return rx
          ll=rx['lst']
          if remote=='yes' and os.path.isfile(fcpuinfo): os.remove(fcpuinfo)

          for q in ll:
              q=q.strip()
              if q!='':
                 x1=q.find(':')
                 if x1>0:
                    k=q[0:x1].strip()
                    v=q[x1+1:].strip()

                    if k=='processor':
                       if not first_skipped:
                          first_skipped=True
                       else:
                          pp+=1
                          spp=str(pp)
                          info_cpu[spp]={}

                    if k!='':
                       info_cpu[spp][k]=v

                       if k.find('MHz')>=0:
                          target_freq[spp]=float(v)
          target_num_proc=str(pp+1)

       # Legacy - outdated - should check name of each CPU (can be different in big-little, for example)
       target_cpu=info_cpu[spp].get('Hardware','')
       if target_cpu=='':
          target_cpu=info_cpu[spp].get('model name','')
       target_sub_cpu=info_cpu[spp].get('Processor','')
       if target_sub_cpu=='':
          target_sub_cpu=info_cpu[spp].get('model name','')
       target_cpu_features=info_cpu[spp].get('Features','')
       if target_cpu_features=='':
          target_cpu_features=info_cpu[spp].get('flags','')

       wa={'device_config':{'core_clusters':[], 'core_names':[]}}
       wa_id=0
       wa_unique={}
       xtcpu=''

       # Process each processor
       for px in range(0, pp+1):
           spx=str(px)

           tcpu=info_cpu[spx].get('Hardware','')
           if tcpu=='':
              tcpu=info_cpu[spx].get('model name','')

           if tcpu=='': # new format
              ic=info_cpu[spx]
              ic1=ic.get('CPU implementer','')
              ic2=ic.get('CPU architecture','')
              ic3=ic.get('CPU variant','')
              ic4=ic.get('CPU part','')
              ic5=ic.get('CPU revision','')

              tcpu=ic1+'-'+ic2+'-'+ic3+'-'+ic4+'-'+ic5

              info_cpu[spx]['cpu_abi']=params.get('ro.product.cpu.abi','') # not sure if correct for all processors
              info_cpu[spx]['new_format']='yes'

              new_format='yes'
           else:
              tsub_cpu=info_cpu[spp].get('Processor','')
              if tsub_cpu=='':
                 tsub_cpu=info_cpu[spp].get('model name','')
              info_cpu[spx]['ck_cpu_subname']=tsub_cpu

           x=params.get('ro.product.cpu.abi','') # not sure if correct for all processors
           if info_cpu[spx].get('cpu_abi','')=='' and x!='':
              info_cpu[spx]['cpu_abi']=x

           info_cpu[spx]['ck_cpu_name']=tcpu

           tcpu_features=info_cpu[spp].get('Features','')
           if tcpu_features=='':
              tcpu_features=info_cpu[spp].get('flags','')
           info_cpu[spx]['cpu_features']=tcpu_features

           # add unique
           found=False
           for uc in unique_cpus:
               if uc.get('ck_cpu_name','')==tcpu:
                  found=True
                  break

           if not found and tcpu!='' and tcpu!='----':
               unique_cpus.append(info_cpu[spx])

           if xtcpu=='' and tcpu!='' and tcpu!='----':
               xtcpu=tcpu

           # add unique in Workload Automation format
           if tcpu!='':
               if tcpu not in wa_unique:
                   wa_unique[tcpu]=wa_id
                   if tcpu!='----':
                       wa_id+=1

               wa_idx=wa_unique[tcpu]

               wa['device_config']['core_names'].append(tcpu)
               wa['device_config']['core_clusters'].append(wa_idx)

       # Clean up CPU names for workload automation
       wa_cn=wa['device_config']['core_names']
       for q in range(0, len(wa_cn)):
           tcpu=wa_cn[q]
           if tcpu=='----' and xtcpu!='':
               wa_cn[q]=xtcpu

       # Collect all frequencies
       for px in range(0, pp+1):
           fnx='/sys/devices/system/cpu/cpu'+str(px)+'/cpufreq/scaling_cur_freq'
           if remote=='yes':
              rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
              if rx['return']>0: return rx
              ffreq=rx['file_name']

              x=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat '+fnx+' '+tosd.get('remote_shell_end','')+' '+ro+ffreq

              if o=='con' and pdv=='yes':
                 ck.out('')
                 ck.out('Executing: '+x)

              rx=os.system(x)
              if rx!=0:
                 if o=='con':
                    ck.out('')
                    ck.out('Non-zero return code :'+str(rx)+' - likely failed')
#                 can fail on MacOS, hence continue ...
#                 return {'return':1, 'error':'access to remote device failed'}
           else:
              ffreq=fnx

           # Read and parse file
           rx=ck.load_text_file({'text_file':ffreq, 'split_to_list':'yes'})
           if rx['return']==0: 
              ll=rx['lst']
              if remote=='yes' and os.path.isfile(ffreq): os.remove(ffreq)

              if len(ll)>0:
                 llx=ll[0].strip()
                 if llx!='':
                    fr=0
                    try:
                      fr=float(llx)/1000
                      target_freq[str(px)]=fr
                      info_cpu[str(px)]['cur_freq']=fr
                    except ValueError:
                      pass

           fnx='/sys/devices/system/cpu/cpu'+str(px)+'/cpufreq/cpuinfo_max_freq'
           if remote=='yes':
              rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
              if rx['return']>0: return rx
              ffreq=rx['file_name']

              x=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat '+fnx+' '+tosd.get('remote_shell_end','')+' '+ro+ffreq

              if o=='con' and pdv=='yes':
                 ck.out('')
                 ck.out('Executing: '+x)

              rx=os.system(x)
              if rx!=0:
                 if o=='con':
                    ck.out('')
                    ck.out('Non-zero return code :'+str(rx)+' - likely failed')
#                 can fail on MacOS, hence continue ...
#                 return {'return':1, 'error':'access to remote device failed'}
           else:
              ffreq=fnx

           # Read and parse file
           rx=ck.load_text_file({'text_file':ffreq, 'split_to_list':'yes'})
           if rx['return']==0:
              ll=rx['lst']
              if remote=='yes' and os.path.isfile(ffreq): os.remove(ffreq)

              if len(ll)>0:
                 llx=ll[0].strip()
                 if llx!='':
                    try:
                      fr=float(llx)/1000
                      target_freq_max[str(px)]=fr
                      info_cpu[str(px)]['max_freq']=fr
                    except ValueError:
                      pass

           fnx='/sys/devices/system/cpu/cpu'+str(px)+'/cpufreq/scaling_available_frequencies'
           if remote=='yes':
              rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
              if rx['return']>0: return rx
              ffreq=rx['file_name']

              x=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat '+fnx+' '+tosd.get('remote_shell_end','')+' '+ro+ffreq

              if o=='con' and pdv=='yes':
                 ck.out('')
                 ck.out('Executing: '+x)

              rx=os.system(x)
              if rx!=0:
                 if o=='con':
                    ck.out('')
                    ck.out('Non-zero return code :'+str(rx)+' - likely failed')
#                 can fail on MacOS, hence continue ...
#                 return {'return':1, 'error':'access to remote device failed'}
           else:
              ffreq=fnx

           # Read and parse file
           rx=ck.load_text_file({'text_file':ffreq, 'split_to_list':'yes'})
           if rx['return']==0:
              ll=rx['lst']
              if remote=='yes' and os.path.isfile(ffreq): os.remove(ffreq)

              if len(ll)>0:
                 llx=ll[0].strip()
                 if llx!='':
                    all=llx.split(' ')
                    for h in all:
                        ppx=str(px)
                        if ppx not in target_freq_all: target_freq_all[ppx]=[]
                        try:
                           h=int(h)
                           target_freq_all[ppx].append(h)
                           if 'all_freqs' not in info_cpu[ppx]:
                              info_cpu[ppx]['all_freqs']=[]
                           info_cpu[ppx]['all_freqs'].append(ppx)
                        except ValueError:
                           pass

# FGG - it is already initalized here!
#       # Initialized device if needed
#       if sdi!='yes':
#          remote_init=tosd.get('remote_init','')
#          if remote_init!='':
#             r=ck.access({'action':'init_device',
#                          'module_uoa':cfg['module_deps']['platform'],
#                          'os_dict':tosd,
#                          'device_id':tdid})
#             if r['return']>0: return r

#       if new_format=='yes':
#          for px in range(0, pp+1):
#              spx=str(px)
#
#              tcpu=info_cpu[spx].get('ck_cpu_name','')
#
#              # add unique
#              found=False
#              for uc in unique_cpus:
#                  if uc.get('ck_cpu_name','')==tcpu:
#                     found=True
#                     break
#
#              if not found and tcpu!='' and tcpu!='----':
#                 unique_cpus.append(info_cpu[spx])

       target['name']=target_cpu
       target['sub_name']=target_sub_cpu
       target['cpu_features']=target_cpu_features
       target['cpu_abi']=params.get('ro.product.cpu.abi','')
       target['num_proc']=target_num_proc
       target['current_freq']=target_freq
       target['max_freq']=target_freq_max
       target['all_freqs']=target_freq_all
       target['workload_automation']=wa

    ########################################################################################
    elif win=='yes':
      r=ck.access({'action':'get_from_wmic',
                   'module_uoa':cfg['module_deps']['platform'],
                   'group':'cpu',
                   'remote_shell':tosd.get('remote_shell','').replace('$#device#$',dv),
                   'remote_shell_end':tosd.get('remote_shell_end','')})
      if r['return']>0: return r
      info_cpu=r['dict']

      target_cpu=info_cpu.get('Name','')

      target_freq=int(info_cpu.get('CurrentClockSpeed','0'))
      target_freq_max=int(info_cpu.get('MaxClockSpeed','0'))

      target_num_proc=int(info_cpu.get('NumberOfLogicalProcessors','0'))

      target['name']=target_cpu
      target['sub_name']=target_cpu
      target['num_proc']=target_num_proc
      target['current_freq']={"0":target_freq}
      target['max_freq']={"0":target_freq_max}

      # To partially support new format
      unique_cpu={'ck_cpu_name':target_cpu,
                  'ck_cpu_subname':target_cpu}
      unique_cpus.append(unique_cpu)

    ########################################################################################
    elif mac=='yes':
      r=ck.access({'action':'run_and_get_stdout',
                   'module_uoa':cfg['module_deps']['os'],
                   'cmd': ['sysctl', 'machdep.cpu', 'hw.cpufrequency']})
      if r['return']>0: return r

      info_cpu={}
      for line in r['stdout'].splitlines():
        if ':' in line:
          left, right = line.split(':', 1)
          left = left.strip().lower()
          right = right.strip()
          info_cpu[left]=right

      target_cpu = info_cpu.get('machdep.cpu.brand_string','')
      target_num_proc = info_cpu.get('machdep.cpu.thread_count','')
      if target_num_proc == '':
        target_num_proc = info_cpu.get('machdep.cpu.core_count','')

      target_freq = int(info_cpu.get('hw.cpufrequency','0')) / 1000000
      target_freq_max = target_freq

      target['name']=target_cpu
      target['sub_name']=target_cpu
      target['cpu_features']=info_cpu.get('machdep.cpu.features','').lower()
      target['cpu_abi']=''
      target['num_proc']=target_num_proc
      target['current_freq']={'0':target_freq}
      target['max_freq']={'0':target_freq_max}

      # To partially support new format
      unique_cpu={'ck_cpu_name':target_cpu,
                  'ck_cpu_subname':target_cpu}
      unique_cpus.append(unique_cpu)      

    ########################################################################################
    if o=='con' and pdv!='no':
       ck.out('')
       if new_format=='yes':
          lup=len(unique_cpus)
          ck.out('Number of logical processors: '+str(target.get('num_proc',0)))
          ck.out('Number of unique processors:  '+str(lup))

          iup=0
          for up in unique_cpus:
              ck.out('')
              ck.out('  Unique processor: '+str(iup))

              x1=up.get('ck_cpu_name','')
              x1x=up.get('ck_cpu_name_real','')
              if x1x!='': x1=x1x+' ('+x1+')'

              x2=up.get('cpu_abi','')
              x3=up.get('cpu_features','')

              ck.out('    CPU name:     '+x1)
              ck.out('    CPU ABI:      '+x2)
              ck.out('    CPU features: '+x3)

              iup+=1

       else:
          ck.out('Number of logical processors: '+str(target.get('num_proc',0)))
          ck.out('CPU name:                     '+target.get('name',''))
          if target.get('name','')!=target.get('sub_name',''):
             ck.out('CPU sub name:                 '+target.get('sub_name',''))
          ck.out('CPU ABI:                      '+target.get('cpu_abi',''))
          ck.out('CPU features:                 '+target.get('cpu_features',''))

       ck.out('')
       ck.out('CPU frequency:')
       x=target.get('current_freq',{})
       for k in sorted(x, key=ck.convert_str_key_to_int):
           v=x[k]
           ck.out('  CPU'+k+' = '+str(v)+' MHz')
       ck.out('CPU max frequency:')
       x=target.get('max_freq',{})
       for k in sorted(x, key=ck.convert_str_key_to_int):
           v=x[k]
           ck.out('  CPU'+k+' = '+str(v)+' MHz')

       x=target.get('all_freqs',{})
       if len(x)>0:
          ck.out('CPU all frequencies (Hz):')
          import json
          for k in sorted(x, key=ck.convert_str_key_to_int):
              v=x[k]
              ck.out('  CPU'+k+' = '+json.dumps(v))

    fuoa=''
    fuid=''

    # Exchanging info #################################################################
    if ex=='yes':
       er=i.get('exchange_repo','')
       esr=i.get('exchange_subrepo','')
       el=i.get('exchange_locally','')
       if el!='yes' and er=='': 
          er=ck.cfg['default_exchange_repo_uoa']
          esr=ck.cfg['default_exchange_subrepo_uoa']

       if new_format=='yes':
          for unique in unique_cpus:
              xn=unique.get('ck_cpu_name','')

              if o=='con':
                 ck.out('')
                 ck.out('Exchanging information with '+er+' repository for a unique processor '+xn+' ...')

              # Copy nearly all (remove cur freq)

              mm={'features':unique}

              ii={'action':'exchange',
                  'module_uoa':cfg['module_deps']['platform'],
                  'sub_module_uoa':work['self_module_uid'],
                  'repo_uoa':er,
                  'data_name':xn,
                  'extra_info':einf,
                  'all':'no',
                  'dict':mm}
              if esr!='': ii['remote_repo_uoa']=esr
              r=ck.access(ii)
              if r['return']>0: return r

              if r.get('found','')=='yes':
                 fuoa=r.get('data_uoa','')
                 fuid=r.get('data_uid','')

                 ddd=r['dict'].get('features',{})

                 if o=='con':
                    ck.out('  CPU CK entry already exists ('+fuid+') - loading latest meta (features) ...')
                    x1=ddd.get('ck_processor_real_name','')
                    x2=ddd.get('ck_arch_real_name','')

                    x=x1
                    if x=='': x=x2
                    else:
                       if x2!='':
                          x+=' ('+x2+')'

                    if x!='':
                       ck.out('    Real name: '+x)

                 ddd['data_uoa']=fuoa
                 ddd['data_uid']=fuid

                 unique.update(ddd)

                 # Update all processors with this name
                 for px in range(0, pp+1):
                     spx=str(px)

                     tcpu=info_cpu[spx].get('ck_cpu_name','')
                     if tcpu==xn:
                        info_cpu[spx].update(ddd)

       else:
           if o=='con':
              ck.out('')
              ck.out('Exchanging information with '+er+' repository (old format) ...')

           xn=target.get('name','')

           if xn=='':
              # Check if exists in configuration
              dcfg={}
              ii={'action':'load',
                  'module_uoa':cfg['module_deps']['cfg'],
                  'data_uoa':cfg['cfg_uoa']}
              r=ck.access(ii)
              if r['return']>0 and r['return']!=16: return r
              if r['return']!=16:
                 dcfg=r['dict']

              dx=dcfg.get('platform_cpu_name',{}).get(tos,{})
              x=tdid
              if x=='': x='default'
              xn=dx.get(x,'')

              if (xn=='' and o=='con'):
                 r=ck.inp({'text':'Enter your processor name: '})
                 xxn=r['string'].strip()

                 if xxn!=xn:
                    xn=xxn

                    if 'platform_cpu_name' not in dcfg: dcfg['platform_cpu_name']={}
                    if tos not in dcfg['platform_cpu_name']: dcfg['platform_cpu_name'][tos]={}
                    dcfg['platform_cpu_name'][tos][x]=xn

                    ii={'action':'update',
                        'module_uoa':cfg['module_deps']['cfg'],
                        'data_uoa':cfg['cfg_uoa'],
                        'dict':dcfg}
                    r=ck.access(ii)
                    if r['return']>0: return r

              if xn=='':
                 return {'return':1, 'error':'can\'t exchange information where main name is empty'}

              target['name']=xn

           # Copy nearly all (remove cur freq)
           import copy
           xtarget=copy.deepcopy(target)
           if 'current_freq' in xtarget: del(xtarget['current_freq'])

           ii={'action':'exchange',
               'module_uoa':cfg['module_deps']['platform'],
               'sub_module_uoa':work['self_module_uid'],
               'repo_uoa':er,
               'data_name':target.get('name',''),
               'extra_info':einf,
               'all':'no',
               'dict':{'features':xtarget}} # Later we should add more properties from prop_all,
                                            # but should be careful to remove any user-specific info
           if esr!='': ii['remote_repo_uoa']=esr
           r=ck.access(ii)
           if r['return']>0: return r

           fuoa=r.get('data_uoa','')
           fuid=r.get('data_uid','')

           eft=r['dict'].get('features',{})

           if o=='con' and r.get('found','')=='yes':
              ck.out('  CPU CK entry already exists ('+fuid+') - loading latest meta (features) ...')
              target=eft

    # Finalize features
    if 'features' not in rr: rr['features']={}

    rr['features']['cpu']=target

    rr['features']['cpu_misc']=info_cpu
    rr['features']['cpu_unique']=unique_cpus
    rr['features']['cpu_new_format']=new_format

    if fuoa!='' or fuid!='':
       rr['features']['cpu_uoa']=fuoa
       rr['features']['cpu_uid']=fuid

    return rr

##############################################################################
# set frequency

def set_freq(i):
    """
    Input:  {
              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)

              (value) = "max" (default)
                        "min"
                        "ondemand"
                        int value

              (env)                  - current env to specialize freq (CK_CPU*)...
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    v=i.get('value','')
    if v=='': v='max'

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    env=i.get('env',{})

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.os']
    ii['skip_info_collection']='yes'
    ii['skip_device_init']='yes'
    rr=ck.access(ii)
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tbits=tosd.get('bits','')

    envtsep=tosd.get('env_separator','')
    etset=tosd.get('env_set','')

    tdid=rr['device_id']

    dir_sep=tosd.get('dir_sep','')

    remote=tosd.get('remote','')
    rs=tosd.get('remote_shell','')
    rse=tosd.get('remote_shell_end','')

    # Prepare scripts
    cmd=''
    if v=='min':
       cmd=tosd.get('script_set_min_cpu_freq','')
    elif v=='max':
       cmd=tosd.get('script_set_max_cpu_freq','')
    elif v=='ondemand':
       cmd=tosd.get('script_set_ondemand_cpu_freq','')
    else:
       cmd=tosd.get('script_set_cpu_freq','').replace('$#freq#$',str(v))

    if cmd!='':
       # Check path to scripts from env
       path_to_scripts=''

       pi_uoa=os.environ.get('CK_PLATFORM_INIT_UOA','')
       if pi_uoa=='':
          dcfg={}
          ii={'action':'load',
              'module_uoa':cfg['module_deps']['cfg'],
              'data_uoa':cfg['lcfg_uoa']}
          r=ck.access(ii)
          if r['return']>0 and r['return']!=16: return r
          if r['return']!=16:
             pi_key=tosx
             if remote=='yes' and tdid!='': pi_key+='-'+tdid

             dcfg=r['dict']
             pi_uoa=dcfg.get('platform_init_uoa',{}).get(pi_key,'')

       if pi_uoa!='' and remote!='yes':
          rx=ck.access({'action':'find',
                        'module_uoa':cfg['module_deps']['platform.init'],
                        'data_uoa':pi_uoa})
          if rx['return']==0:
             path_to_scripts=rx['path']

       if path_to_scripts=='':
          path_to_scripts=tosd.get('path_to_scripts','')


       if path_to_scripts!='':
          cmd=path_to_scripts+dir_sep+cmd

       # Add env
       xcmd=''
       for k in sorted(env):
           if k.startswith('CK_CPU'):
              v=str(env[k])

              if xcmd!='': xcmd+=envtsep
              xcmd+=etset+' '+k+'='+v

       if xcmd!='': 
          cmd=xcmd+envtsep+cmd

       if o=='con':
          ck.out('')
          ck.out('CMD to set CPU frequency:')
          ck.out('  '+cmd)

       # Get all params
       if remote=='yes':
          dv=''
          if tdid!='': dv=' -s '+tdid

          x=rs.replace('$#device#$',dv)
          # Support backwards compatibility if double quote not found in rs and rse
          x+=' '+('"' if rs.find('"') == -1 else '')+' '+cmd+' '+rse+('"' if rse.find('"') == -1 else '')

          rx=os.system(x)
          if rx!=0:
             if o=='con':
                ck.out('')
                ck.out('Non-zero return code :'+str(rx)+' - likely failed')

       else:
             rx=os.system(cmd)
             if rx!=0:
                if o=='con':
                   ck.out('')
                   ck.out('  Warning: setting frequency possibly failed - return code '+str(rx))

    return {'return':0}

##############################################################################
# viewing entries as html

def show(i):
    """
    Input:  {
              data_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """


    h='<h2>Processors of platforms participating in crowd-tuning</h2>\n'

    h+='<i>Reuse/extend <a href="https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.cpu">CK JSON meta information</a> of these processors using "ck pull repo:ck-crowdtuning-platforms" ...</i><br><br>\n'

    h+='<table class="ck_table" border="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   #\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Name 1\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Name 2\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Cores\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Max frequency (MHz)\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   ABI\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Features\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':">CK UID</a>\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    ruoa=i.get('repo_uoa','')
    muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_info':'yes',
                 'add_meta':'yes'})
    if r['return']>0: 
       return {'return':0, 'html':'Error: '+r['error']}

    lst=r['lst']

    num=0
    for q in sorted(lst, key = lambda x: (x.get('meta',{}).get('features',{}).get('sub_name','').upper(), \
                                          x.get('meta',{}).get('features',{}).get('name','').upper())):

        num+=1

        duoa=q['data_uoa']
        duid=q['data_uid']

        meta=q['meta']
        ft=meta.get('features',{})

        name=ft.get('name','')
        if name=='':
           name=ft.get('ck_arch_real_name','')

        sub_name=ft.get('sub_name','')

        cores=ft.get('num_proc','')
        abi=ft.get('cpu_abi','')
        features=ft.get('cpu_features','')
        frequency=ft.get('max_freq',{})

        freq=''
        for x in range(0,1024):
            xx=str(x)
            if type(frequency)!=list or xx not in frequency: break
            if freq!='': freq+=', '
            freq+=str(frequency[xx])

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(num)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+sub_name+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+name+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(cores)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+freq+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+abi+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+features+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':'+duid+'">'+duid+'</a>\n'
        h+='  </td>\n'
        h+=' </tr>\n'


    h+='</table><br><br>\n'

    return {'return':0, 'html':h}

##############################################################################
# browse platform.cpus participated in experiment crowdsourcing (crowd-benchmarking and crowd-tuning)

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # TBD: should calculate url
    url='https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.cpu'

    import webbrowser
    webbrowser.open(url)

    import time
    time.sleep(3)

    url='http://cknowledge.org/repo/web.php?action=index&module_uoa=wfe&native_action=show&native_module_uoa=platform.cpu'

    import webbrowser
    webbrowser.open(url)

    return {'return':0}



#
# Collective Knowledge (pipeline demo)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# run pipeline

def pipeline(i):
    """
    Input:  {
               (cmd)            - cmd string or list
               (compiler_vars)  - will substitute dummies $#VAR#$ in cmd
               (compiler_flags) - assemble into string and substitute $#compiler_flags#$
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import json
    import os
    import time
    import sys

    o=i.get('out','')

    cmd=i.get('cmd','')

    if type(cmd)!=list:
       cmd=[cmd]

    # Checking some known vars 
    cv=i.get('compiler_vars',{})
    cf=i.get('compiler_flags',{})

    cflags=''
    for c in cf:
        cx='##compiler_flags#'+c
        cc=cf[c]
        if cc!='':
           if cflags!='': cflags+=' '
           cflags+=str(cc)

    # Parsing special vars $#var#$
    pcmd=[]
    for c in cmd:
        j=c.find('$#')
        while j>=0:
           j1=c.find('#$',j+1)
           if j1<0:
              return {'return':1, 'error':'inconsistent CK vars $# #$ in cmd ('+c+')'}

           k=c[j+2:j1]
 
           # Check special keys
           if k=='cflags':
              v=cflags
           else:
              v=cv.get(k,'') # for backwards compatibility 
              if v=='':
                 v=i.get(k,'')

           c=c[:j]+str(v)+c[j1+2:]

           j=c.find('$#')
           
        pcmd.append(c)

    # Print

    if o=='con':
       ck.out('Updated command line in the pipeline:')
       ck.out('')
       for c in pcmd:
           ck.out('  '+c)

       ck.out('')
       ck.out('Executing ...')
       ck.out('')

    start_time=time.time()

    sys.stdout.flush()

    for c in pcmd:
        os.system(c.strip())

    run_time=time.time()-start_time

    if o=='con':
       ck.out('')
       ck.out('Execution time: '+('%.3f'%run_time)+' sec.')
       ck.out('')


    return {'return':0, 'run_time':run_time}



#
# Collective Knowledge (experiment scenarios to be executed on Android during crowdsourcing)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# get scenarios depending on user's mobile device features

def get(i):
    """
    Input:  {
              (data_uoa)
              (repo_uoa)

              (platform_features)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              scenarios    - list of scenarios and related files
            }

    """

    import copy

    pf=i.get('platform_features',{})

    abi=pf.get('cpu',{}).get('cpu_abi','')

    os_name=pf.get('os',{}).get('name','')
    os_ver=[]
    j=os_name.find(' ')
    if j>0:
        os_ver=os_name[j+1:].strip().split('.')

    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_meta':'yes'})
    if r['return']>0: return r

    lst=r['lst']

    nlst=[]

    # Prepare URL from CK server
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    # Calculate total size
    for q in sorted(lst, key=lambda x: x.get('meta',{}).get('title','')):
        add=True

        meta=q['meta']

        if meta.get('skip','')!='yes' and meta.get('outdated','')!='yes':
            sabi=meta.get('supported_abi',[])
            if abi!='' and abi not in sabi:
                add=False

            if add:
                min_os_ver=meta.get('min_os_ver',[])
                # TBD: need to check all digits
                if len(min_os_ver)>0 and len(os_ver)>0 and os_ver[0]<min_os_ver[0]:
                    add=False

            if add:
                # Check engine_meta ->
                em=meta.get('engine_meta',{}).get(abi,{})
                if em>0:
                   meta['engine_meta_dict']=copy.deepcopy(em)
                   del(meta['engine_meta'])

                ff=meta.get('files',[])

                # Go through files and update
                nff=[]
                tfs=0 # Total file size

                for f in ff:
                    sabi=f.get('supported_abi',[])
                    if len(sabi)==0 or abi=='' or abi in sabi:
                        url=f.get('url','')
                        if url=='':
                            path=f.get('path','')
                            fn=f.get('filename','')

                            dduoa=q['data_uid']
                            if f.get('from_data_uoa','')!='':
                                dduoa=f['from_data_uoa']

                            url=url0+'action=pull&common_action=yes&cid='+q['module_uoa']+':'+dduoa+'&filename='+path+'/'+fn

                        f['url']=url
                        nff.append(f)

                        fs=int(f.get('file_size',0))
                        tfs+=fs

                q['total_file_size']=tfs

                tfs=int(tfs/1E6)

                if tfs>0:
                    title=meta.get('title','')
                    if title!='':
#                        title+=' ('+str(tfs+1)+' MB)'
                        meta['title']=title

                meta['files']=nff

                nlst.append(q)

    # Sort by title
    snlst=nlst
#[]
#    for q in sorted(nlst, key=lambda x: (x.get('total_file_size',0),x.get('meta',{}).get('title',''))):
#    for q in sorted(nlst, key=lambda x: x.get('meta',{}).get('title','')):
#        snlst.append(q)

#    ck.save_json_to_file({'json_file':'/tmp/xyz888.json','dict':snlst})

    return {'return':0, 'scenarios':snlst}

##############################################################################
# process all scenarios (check files, get md5, get length)

def process(i):
    """
    Input:  {
              (data_uoa)
              (repo_uoa)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_meta':'yes'})
    if r['return']>0: return r

    lst=r['lst']

    nlst=[]

    # Prepare URL from CK server
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    for q in lst:
        meta=q['meta']

        if meta.get('outdated','')=='yes':
           continue

        ck.out('Processing '+q['data_uoa']+' ...')

        ff=meta.get('files',[])

        # Go through files and update
        for f in ff:
            fn=f.get('filename','')
            ck.out('  '+fn)

            url=f.get('url','')
            if url=='':
                p=f.get('path','')

                if not p.startswith('code') and not p.startswith('data'):
                   j1=p.find('code/')
                   if j1<0:
                      j1=p.find('data/')
                   if j1>0:
                      p=p[j1:]
                      f['path']=p

                md5x=f.get('md5','')

                dduoa=q['data_uid']
                if f.get('from_data_uoa','')!='':
                   dduoa=f['from_data_uoa']

                # Get path
                r=ck.access({'action':'load',
                             'data_uoa':dduoa,
                             'module_uoa':q['module_uoa']})
                if r['return']>0: return r

                pp=os.path.join(r['path'],p,fn)

                if not os.path.isfile(pp):
                    return {'return':1, 'error':'file ('+pp+') not found'}

                # Check MD5 sum
                md5=''

                r=ck.run_and_get_stdout({'cmd':'md5sum < '+pp})
                if r['return']>0: return r

                s=r['stdout'].split(' ')
                if len(s)>0:
                    md5=s[0]
                    if len(md5)>0 and md5.startswith('\\'):
                        md5=md5[1:]

                f['md5']=md5

                # Get file size
                fs=os.path.getsize(pp)
                f['file_size']=fs

                ck.out('    File size = '+str(fs))
                ck.out('    MD5       = '+md5)

        # Update entry
        r=ck.access({'action':'update',
                     'module_uoa':q['module_uid'],
                     'data_uoa':q['data_uid'],
                     'dict':meta,
                     'substitute':'yes',
                     'sort_key':'yes',
                     'ignore_update':'yes'})
        if r['return']>0: return r

    return {'return':0, 'scenarios':nlst}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

import unittest

ck=None # Will be updated by CK (initialized CK kernel)

class CompareVersionsTests(unittest.TestCase):
    def test_get_version(self):
        """
            expected output format:
                ret_struct = {'version': ['1', '9', '4', '1'], 'return': 0, 'version_str': '1.9.4.1'}
        """

        ret_struct = ck.version( {} )
        self.assertEqual(   ret_struct['return'],               0,                          ret_struct.get('error', None) )
        self.assertEqual(   type(ret_struct['version']),        list,                       '["version"] should be a list' )
        self.assertGreater( len(ret_struct['version']),         0,                          '["version"] should contain at least one element' )
        self.assertEqual(   '.'.join(ret_struct['version']),    ret_struct['version_str'],  '["version"] is correctly parsed from ["version_str"]')

    def test_compare_versions(self):
        """
            expected output format:
                ret_struct = {'return': 0, 'result': '<'}
        """

        def compare_one_pair_one_way(left_version, right_version, expected_sign):
            "Run one verbose comparison"

            ret_struct = ck.access({
                'action':       'compare_versions',
                'module_uoa':   'soft',
                'version1':     left_version,
                'version2':     right_version
            })
            result = ret_struct['result']
            self.assertEqual(   ret_struct['return'],   0,          ret_struct.get('error', None) )
            outcome_text = ('(ok)' if expected_sign==result else '(expected '+expected_sign+')') if expected_sign else ''
            print('compare_versions:   %12s   %s  %12s      %s' % (str(left_version),result, str(right_version), outcome_text) )

            if expected_sign:
                self.assertEqual(   result,                 expected_sign,   "Expected '"+expected_sign+"', but got '"+result+"' instead")

        def compare_direct_and_reverse(left_version, right_version, expected_sign=None):
            "Try both directions to test for symmetry of < / > relation"

            compare_one_pair_one_way(left_version, right_version, expected_sign)
            compare_one_pair_one_way(right_version, left_version, {'>':'<', '<':'>', '=':'=', None:None}[expected_sign] )


        # Your favourite vectors go here:
        #
        ver_10ab    = [ 1, 0, 'ab' ]
        ver_103     = [ 1, 0, 3 ]
        ver_12      = [ 1, 2 ]
        ver_120     = [ 1, 2, 0 ]
        ver_125     = [ 1, 2, 5 ]
        ver_13      = [ 1, 3 ]
        ver_130     = [ 1, 3, 0 ]
        ver_2107    = [ 2, 10, 7 ]
        ver_2_27    = [ 2,  2, 7 ]

        compare_direct_and_reverse(ver_12,      ver_120,    '=')    # switched to new behaviour
        compare_direct_and_reverse(ver_120,     ver_120,    '=')
        compare_direct_and_reverse(ver_125,     ver_120,    '>')
        compare_direct_and_reverse(ver_103,     ver_13,     '<')
        compare_direct_and_reverse(ver_103,     ver_130,    '<')    # against an old bug where vector components were sorted first
        compare_direct_and_reverse(ver_103,     ver_10ab,   '>')    # any integer is higher than any letter combo
        compare_direct_and_reverse(ver_2107,    ver_2_27,   '>')    # against alphabetic ordering of integers

if __name__ == '__main__':
    import ck.kernel
    ck = ck.kernel
    unittest.main()



#
# Collective Knowledge (CK)
#
# See CK LICENSE.txt for licensing details.
# See CK COPYRIGHT.txt for copyright details.
#
# Developer: Grigori Fursin
#

# CK dummy

import sys
import ck.kernel as ck

r = ck.access(sys.argv[1:])

if 'return' not in r:
    raise Exception('CK access function should always return key \'return\'!')

exit(int(r['return']))
